from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from pydantic import Field
from pydantic import field_validator, model_validator, ValidationInfo

from cognite.powerops.client._generated.v1.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    T_DomainModelList,
    as_node_id,
    as_read_args,
    as_write_args,
    is_tuple_id,
    as_instance_dict_id,
    parse_single_connection,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    ViewPropertyId,
    DirectRelationFilter,
    IntFilter,
)
from cognite.powerops.client._generated.v1.data_classes._function_output import FunctionOutput, FunctionOutputWrite
if TYPE_CHECKING:
    from cognite.powerops.client._generated.v1.data_classes._alert import Alert, AlertList, AlertGraphQL, AlertWrite, AlertWriteList
    from cognite.powerops.client._generated.v1.data_classes._bid_configuration_day_ahead import BidConfigurationDayAhead, BidConfigurationDayAheadList, BidConfigurationDayAheadGraphQL, BidConfigurationDayAheadWrite, BidConfigurationDayAheadWriteList
    from cognite.powerops.client._generated.v1.data_classes._bid_matrix import BidMatrix, BidMatrixList, BidMatrixGraphQL, BidMatrixWrite, BidMatrixWriteList
    from cognite.powerops.client._generated.v1.data_classes._partial_bid_matrix_calculation_input import PartialBidMatrixCalculationInput, PartialBidMatrixCalculationInputList, PartialBidMatrixCalculationInputGraphQL, PartialBidMatrixCalculationInputWrite, PartialBidMatrixCalculationInputWriteList


__all__ = [
    "PartialBidMatrixCalculationOutput",
    "PartialBidMatrixCalculationOutputWrite",
    "PartialBidMatrixCalculationOutputApply",
    "PartialBidMatrixCalculationOutputList",
    "PartialBidMatrixCalculationOutputWriteList",
    "PartialBidMatrixCalculationOutputApplyList",
    "PartialBidMatrixCalculationOutputFields",
    "PartialBidMatrixCalculationOutputTextFields",
    "PartialBidMatrixCalculationOutputGraphQL",
]


PartialBidMatrixCalculationOutputTextFields = Literal["external_id", "workflow_execution_id", "function_name", "function_call_id"]
PartialBidMatrixCalculationOutputFields = Literal["external_id", "workflow_execution_id", "workflow_step", "function_name", "function_call_id"]

_PARTIALBIDMATRIXCALCULATIONOUTPUT_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "workflow_execution_id": "workflowExecutionId",
    "workflow_step": "workflowStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
}


class PartialBidMatrixCalculationOutputGraphQL(GraphQLCore):
    """This represents the reading version of partial bid matrix calculation output, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid matrix calculation output.
        data_record: The data record of the partial bid matrix calculation output node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        function_input: The function input field.
        alerts: An array of calculation level Alerts.
        partial_matrix: The partial matrix field.
        bid_configuration: The bid configuration field.
    """

    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PartialBidMatrixCalculationOutput", "1")
    workflow_execution_id: Optional[str] = Field(None, alias="workflowExecutionId")
    workflow_step: Optional[int] = Field(None, alias="workflowStep")
    function_name: Optional[str] = Field(None, alias="functionName")
    function_call_id: Optional[str] = Field(None, alias="functionCallId")
    function_input: Optional[PartialBidMatrixCalculationInputGraphQL] = Field(default=None, repr=False, alias="functionInput")
    alerts: Optional[list[AlertGraphQL]] = Field(default=None, repr=False)
    partial_matrix: Optional[BidMatrixGraphQL] = Field(default=None, repr=False, alias="partialMatrix")
    bid_configuration: Optional[BidConfigurationDayAheadGraphQL] = Field(default=None, repr=False, alias="bidConfiguration")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values


    @field_validator("function_input", "alerts", "partial_matrix", "bid_configuration", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> PartialBidMatrixCalculationOutput:
        """Convert this GraphQL format of partial bid matrix calculation output to the reading format."""
        return PartialBidMatrixCalculationOutput.model_validate(as_read_args(self))

    def as_write(self) -> PartialBidMatrixCalculationOutputWrite:
        """Convert this GraphQL format of partial bid matrix calculation output to the writing format."""
        return PartialBidMatrixCalculationOutputWrite.model_validate(as_write_args(self))


class PartialBidMatrixCalculationOutput(FunctionOutput):
    """This represents the reading version of partial bid matrix calculation output.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid matrix calculation output.
        data_record: The data record of the partial bid matrix calculation output node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        function_input: The function input field.
        alerts: An array of calculation level Alerts.
        partial_matrix: The partial matrix field.
        bid_configuration: The bid configuration field.
    """

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PartialBidMatrixCalculationOutput", "1")

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "PartialBidMatrixCalculationOutput")
    partial_matrix: Union[BidMatrix, str, dm.NodeId, None] = Field(default=None, repr=False, alias="partialMatrix")
    bid_configuration: Union[BidConfigurationDayAhead, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidConfiguration")
    @field_validator("function_input", "partial_matrix", "bid_configuration", mode="before")
    @classmethod
    def parse_single(cls, value: Any, info: ValidationInfo) -> Any:
        return parse_single_connection(value, info.field_name)

    @field_validator("alerts", mode="before")
    @classmethod
    def parse_list(cls, value: Any, info: ValidationInfo) -> Any:
        if value is None:
            return None
        return [parse_single_connection(item, info.field_name) for item in value]

    def as_write(self) -> PartialBidMatrixCalculationOutputWrite:
        """Convert this read version of partial bid matrix calculation output to the writing version."""
        return PartialBidMatrixCalculationOutputWrite.model_validate(as_write_args(self))

    def as_apply(self) -> PartialBidMatrixCalculationOutputWrite:
        """Convert this read version of partial bid matrix calculation output to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PartialBidMatrixCalculationOutputWrite(FunctionOutputWrite):
    """This represents the writing version of partial bid matrix calculation output.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid matrix calculation output.
        data_record: The data record of the partial bid matrix calculation output node.
        workflow_execution_id: The process associated with the function execution
        workflow_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        function_input: The function input field.
        alerts: An array of calculation level Alerts.
        partial_matrix: The partial matrix field.
        bid_configuration: The bid configuration field.
    """
    _container_fields: ClassVar[tuple[str, ...]] = ("bid_configuration", "function_call_id", "function_input", "function_name", "partial_matrix", "workflow_execution_id", "workflow_step",)
    _outwards_edges: ClassVar[tuple[tuple[str, dm.DirectRelationReference], ...]] = (("alerts", dm.DirectRelationReference("power_ops_types", "calculationIssue")),)
    _direct_relations: ClassVar[tuple[str, ...]] = ("bid_configuration", "function_input", "partial_matrix",)

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PartialBidMatrixCalculationOutput", "1")

    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = dm.DirectRelationReference("power_ops_types", "PartialBidMatrixCalculationOutput")
    partial_matrix: Union[BidMatrixWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="partialMatrix")
    bid_configuration: Union[BidConfigurationDayAheadWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="bidConfiguration")

    @field_validator("partial_matrix", "bid_configuration", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value


class PartialBidMatrixCalculationOutputApply(PartialBidMatrixCalculationOutputWrite):
    def __new__(cls, *args, **kwargs) -> PartialBidMatrixCalculationOutputApply:
        warnings.warn(
            "PartialBidMatrixCalculationOutputApply is deprecated and will be removed in v1.0. "
            "Use PartialBidMatrixCalculationOutputWrite instead. "
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PartialBidMatrixCalculationOutput.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)

class PartialBidMatrixCalculationOutputList(DomainModelList[PartialBidMatrixCalculationOutput]):
    """List of partial bid matrix calculation outputs in the read version."""

    _INSTANCE = PartialBidMatrixCalculationOutput
    def as_write(self) -> PartialBidMatrixCalculationOutputWriteList:
        """Convert these read versions of partial bid matrix calculation output to the writing versions."""
        return PartialBidMatrixCalculationOutputWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PartialBidMatrixCalculationOutputWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()

    @property
    def function_input(self) -> PartialBidMatrixCalculationInputList:
        from ._partial_bid_matrix_calculation_input import PartialBidMatrixCalculationInput, PartialBidMatrixCalculationInputList
        return PartialBidMatrixCalculationInputList([item.function_input for item in self.data if isinstance(item.function_input, PartialBidMatrixCalculationInput)])
    @property
    def alerts(self) -> AlertList:
        from ._alert import Alert, AlertList
        return AlertList([item for items in self.data for item in items.alerts or [] if isinstance(item, Alert)])

    @property
    def partial_matrix(self) -> BidMatrixList:
        from ._bid_matrix import BidMatrix, BidMatrixList
        return BidMatrixList([item.partial_matrix for item in self.data if isinstance(item.partial_matrix, BidMatrix)])
    @property
    def bid_configuration(self) -> BidConfigurationDayAheadList:
        from ._bid_configuration_day_ahead import BidConfigurationDayAhead, BidConfigurationDayAheadList
        return BidConfigurationDayAheadList([item.bid_configuration for item in self.data if isinstance(item.bid_configuration, BidConfigurationDayAhead)])

class PartialBidMatrixCalculationOutputWriteList(DomainModelWriteList[PartialBidMatrixCalculationOutputWrite]):
    """List of partial bid matrix calculation outputs in the writing version."""

    _INSTANCE = PartialBidMatrixCalculationOutputWrite
    @property
    def function_input(self) -> PartialBidMatrixCalculationInputWriteList:
        from ._partial_bid_matrix_calculation_input import PartialBidMatrixCalculationInputWrite, PartialBidMatrixCalculationInputWriteList
        return PartialBidMatrixCalculationInputWriteList([item.function_input for item in self.data if isinstance(item.function_input, PartialBidMatrixCalculationInputWrite)])
    @property
    def alerts(self) -> AlertWriteList:
        from ._alert import AlertWrite, AlertWriteList
        return AlertWriteList([item for items in self.data for item in items.alerts or [] if isinstance(item, AlertWrite)])

    @property
    def partial_matrix(self) -> BidMatrixWriteList:
        from ._bid_matrix import BidMatrixWrite, BidMatrixWriteList
        return BidMatrixWriteList([item.partial_matrix for item in self.data if isinstance(item.partial_matrix, BidMatrixWrite)])
    @property
    def bid_configuration(self) -> BidConfigurationDayAheadWriteList:
        from ._bid_configuration_day_ahead import BidConfigurationDayAheadWrite, BidConfigurationDayAheadWriteList
        return BidConfigurationDayAheadWriteList([item.bid_configuration for item in self.data if isinstance(item.bid_configuration, BidConfigurationDayAheadWrite)])

class PartialBidMatrixCalculationOutputApplyList(PartialBidMatrixCalculationOutputWriteList): ...


def _create_partial_bid_matrix_calculation_output_filter(
    view_id: dm.ViewId,
    workflow_execution_id: str | list[str] | None = None,
    workflow_execution_id_prefix: str | None = None,
    min_workflow_step: int | None = None,
    max_workflow_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    function_input: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    partial_matrix: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    bid_configuration: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(workflow_execution_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id))
    if workflow_execution_id and isinstance(workflow_execution_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("workflowExecutionId"), values=workflow_execution_id))
    if workflow_execution_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("workflowExecutionId"), value=workflow_execution_id_prefix))
    if min_workflow_step is not None or max_workflow_step is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("workflowStep"), gte=min_workflow_step, lte=max_workflow_step))
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if isinstance(function_input, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(function_input):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionInput"), value=as_instance_dict_id(function_input)))
    if function_input and isinstance(function_input, Sequence) and not isinstance(function_input, str) and not is_tuple_id(function_input):
        filters.append(dm.filters.In(view_id.as_property_ref("functionInput"), values=[as_instance_dict_id(item) for item in function_input]))
    if isinstance(partial_matrix, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(partial_matrix):
        filters.append(dm.filters.Equals(view_id.as_property_ref("partialMatrix"), value=as_instance_dict_id(partial_matrix)))
    if partial_matrix and isinstance(partial_matrix, Sequence) and not isinstance(partial_matrix, str) and not is_tuple_id(partial_matrix):
        filters.append(dm.filters.In(view_id.as_property_ref("partialMatrix"), values=[as_instance_dict_id(item) for item in partial_matrix]))
    if isinstance(bid_configuration, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(bid_configuration):
        filters.append(dm.filters.Equals(view_id.as_property_ref("bidConfiguration"), value=as_instance_dict_id(bid_configuration)))
    if bid_configuration and isinstance(bid_configuration, Sequence) and not isinstance(bid_configuration, str) and not is_tuple_id(bid_configuration):
        filters.append(dm.filters.In(view_id.as_property_ref("bidConfiguration"), values=[as_instance_dict_id(item) for item in bid_configuration]))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _PartialBidMatrixCalculationOutputQuery(NodeQueryCore[T_DomainModelList, PartialBidMatrixCalculationOutputList]):
    _view_id = PartialBidMatrixCalculationOutput._view_id
    _result_cls = PartialBidMatrixCalculationOutput
    _result_list_cls_end = PartialBidMatrixCalculationOutputList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.ResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_property: ViewPropertyId | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.ResultSetExpression | None = None,
    ):
        from ._alert import _AlertQuery
        from ._bid_configuration_day_ahead import _BidConfigurationDayAheadQuery
        from ._bid_matrix import _BidMatrixQuery
        from ._partial_bid_matrix_calculation_input import _PartialBidMatrixCalculationInputQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_property,
            connection_type,
            reverse_expression,
        )

        if _PartialBidMatrixCalculationInputQuery not in created_types:
            self.function_input = _PartialBidMatrixCalculationInputQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("functionInput"),
                    direction="outwards",
                ),
                connection_name="function_input",
                connection_property=ViewPropertyId(self._view_id, "functionInput"),
            )

        if _AlertQuery not in created_types:
            self.alerts = _AlertQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="alerts",
                connection_property=ViewPropertyId(self._view_id, "alerts"),
            )

        if _BidMatrixQuery not in created_types:
            self.partial_matrix = _BidMatrixQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("partialMatrix"),
                    direction="outwards",
                ),
                connection_name="partial_matrix",
                connection_property=ViewPropertyId(self._view_id, "partialMatrix"),
            )

        if _BidConfigurationDayAheadQuery not in created_types:
            self.bid_configuration = _BidConfigurationDayAheadQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("bidConfiguration"),
                    direction="outwards",
                ),
                connection_name="bid_configuration",
                connection_property=ViewPropertyId(self._view_id, "bidConfiguration"),
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.workflow_execution_id = StringFilter(self, self._view_id.as_property_ref("workflowExecutionId"))
        self.workflow_step = IntFilter(self, self._view_id.as_property_ref("workflowStep"))
        self.function_name = StringFilter(self, self._view_id.as_property_ref("functionName"))
        self.function_call_id = StringFilter(self, self._view_id.as_property_ref("functionCallId"))
        self.function_input_filter = DirectRelationFilter(self, self._view_id.as_property_ref("functionInput"))
        self.partial_matrix_filter = DirectRelationFilter(self, self._view_id.as_property_ref("partialMatrix"))
        self.bid_configuration_filter = DirectRelationFilter(self, self._view_id.as_property_ref("bidConfiguration"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.workflow_execution_id,
            self.workflow_step,
            self.function_name,
            self.function_call_id,
            self.function_input_filter,
            self.partial_matrix_filter,
            self.bid_configuration_filter,
        ])

    def list_partial_bid_matrix_calculation_output(self, limit: int = DEFAULT_QUERY_LIMIT) -> PartialBidMatrixCalculationOutputList:
        return self._list(limit=limit)


class PartialBidMatrixCalculationOutputQuery(_PartialBidMatrixCalculationOutputQuery[PartialBidMatrixCalculationOutputList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, PartialBidMatrixCalculationOutputList)



from __future__ import annotations

from collections.abc import Sequence

from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.v1._api._core import DEFAULT_LIMIT_READ, EdgeAPI, _create_edge_filter
from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE


class BenchmarkingTaskDispatcherOutputDayAheadBenchmarkingSubTasksAPI(EdgeAPI):
    def list(
        self,
        from_benchmarking_task_dispatcher_output_day_ahead: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        from_benchmarking_task_dispatcher_output_day_ahead_space: str = DEFAULT_INSTANCE_SPACE,
        to_function_input: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
        to_function_input_space: str = DEFAULT_INSTANCE_SPACE,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit=DEFAULT_LIMIT_READ,
    ) -> dm.EdgeList:
        """List benchmarking sub task edges of a benchmarking task dispatcher output day ahead.

        Args:
            from_benchmarking_task_dispatcher_output_day_ahead: ID of the source benchmarking task dispatcher output day ahead.
            from_benchmarking_task_dispatcher_output_day_ahead_space: Location of the benchmarking task dispatcher output day aheads.
            to_function_input: ID of the target function input.
            to_function_input_space: Location of the function inputs.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of benchmarking sub task edges to return. Defaults to 25.
                Set to -1, float("inf") or None to return all items.

        Returns:
            The requested benchmarking sub task edges.

        Examples:

            List 5 benchmarking sub task edges connected to "my_benchmarking_task_dispatcher_output_day_ahead":

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> benchmarking_task_dispatcher_output_day_ahead = client.benchmarking_task_dispatcher_output_day_ahead.benchmarking_sub_tasks_edge.list(
                ...     "my_benchmarking_task_dispatcher_output_day_ahead", limit=5
                ... )

        """
        filter_ = _create_edge_filter(
            dm.DirectRelationReference("power_ops_types", "benchmarkingSubTasks"),
            from_benchmarking_task_dispatcher_output_day_ahead,
            from_benchmarking_task_dispatcher_output_day_ahead_space,
            to_function_input,
            to_function_input_space,
            external_id_prefix,
            space,
        )
        return self._list(filter_=filter_, limit=limit)



# Mypy does not understand the pydantic classes that allows both alias and name to be used in population
# https://github.com/pydantic/pydantic/issues/3923
# mypy: disable-error-code="call-arg"
from __future__ import annotations

import logging
import re
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Optional

import yaml
from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import (
    Boolean,
    DirectRelation,
    FileReference,
    Float32,
    Float64,
    Int32,
    Int64,
    Json,
    MappedProperty,
    MultiEdgeConnection,
    Text,
    TimeSeriesReference,
)
from pydantic import ValidationError

import cognite.powerops.client._generated.v1.data_classes as v1_data_classes
from cognite.powerops.resync.data_classes import (
    DataModelConfiguration,
    PropertyConfiguration,
)
from cognite.powerops.resync.utils import check_input_keys, ext_id_factory, parse_external_ids
from cognite.powerops.utils.serialization import load_yaml

logger = logging.getLogger("resync")

__all__ = ["ResyncImporter"]


class ResyncImporter:
    """Importer class for resync to populate the data model from yaml configuration files.

    Attributes:
        working_directory: Path to the working directory.
        toolkit_directory: Path to the toolkit directory.
        data_model_classes: A dictionary of types to DomainModelWrite classes detailing all data model classes to be
                            used for the resync configuration.
        data_model_configuration: A dictionary of types to DataModelConfiguration detailing all configuration
                                  properties of the data model type.
        ignore_nones: A boolean to ignore None values in the data model objects.
        instance_space: A string of the instance space to be used for the data model objects.
        models_space: A string of the models space to be used for the data model objects.
        type_space: A string of the type space to be used for the data model objects.
    """

    working_directory: Path
    toolkit_directory: Path
    folders_to_process: list[Path]
    data_model_classes: dict[str, type]
    data_model_configuration: dict[str, DataModelConfiguration]
    ignore_nones: bool = True
    instance_space: str = "power_ops_instances"
    models_space: str = "power_ops_core"
    type_space: str = "power_ops_types"
    data_model_version: str = "1"

    def __init__(
        self,
        working_directory: Path,
        toolkit_directory: Path,
        data_model_classes: dict[str, type],
        data_model_configuration: Optional[dict[str, DataModelConfiguration]] = None,
        ignore_nones: bool = True,
        instance_space: str = "power_ops_instances",
        models_space: str = "power_ops_core",
        type_space: str = "power_ops_types",
        data_model_version: str = "1",
    ) -> None:
        """Initializes the ResyncImporter"""

        self.working_directory = working_directory
        self.toolkit_directory = toolkit_directory
        self.folders_to_process = [p for p in self.working_directory.iterdir() if p.is_dir()]

        self.data_model_configuration = data_model_configuration or {}
        self.data_model_classes = data_model_classes

        self.ignore_nones = ignore_nones

        self.instance_space = instance_space
        self.models_space = models_space
        self.type_space = type_space
        self.data_model_version = data_model_version

        v1_data_classes.DomainModelWrite.external_id_factory = ext_id_factory

    @classmethod
    def from_yaml(
        cls,
        configuration_path: Path,
        data_model_classes: dict[str, type],
        cdf_client: CogniteClient,
    ) -> ResyncImporter:
        """Creates a ResyncImporter object from a resync configuration file.

        Args:
            configuration_path: Path to the resync configuration file.
            data_model_classes: A dictionary of all data model classes to be used for the resync configuration.

        Returns:
            A ResyncImporter object.
        """
        configuration = load_yaml(configuration_path, expected_return_type="dict")

        working_directory = Path(configuration.get("working_directory", Path.cwd()))

        temp_toolkit_directory = configuration.get("toolkit_directory")
        if temp_toolkit_directory:
            toolkit_directory = Path(temp_toolkit_directory)
        else:
            raise ValueError("toolkit_directory is required in the configuration file")

        return cls(
            working_directory=working_directory,
            toolkit_directory=toolkit_directory,
            data_model_classes=data_model_classes,
            ignore_nones=configuration.get("ignore_nones", True),
            instance_space=configuration.get("instance_space", "power_ops_instances"),
            models_space=configuration.get("models_space", "power_ops_core"),
            type_space=configuration.get("type_space", "power_ops_types"),
            data_model_version=configuration.get("data_model_version", "1"),
        )

    def to_toolkit_nodes_edges(self, client: CogniteClient) -> tuple[list[v1_data_classes.DomainModelWrite], list[str]]:
        """Converts configuration files to toolkit format files for relevant nodes and edges.

        Retrieves raw data from files and using the resync configuration parses the data into
        pygen data model objects which are then dumped and formatted into dictionary yaml toolkit format.

        Returns:
            data_model_objects: A list of all data model objects created from the configuration files.
            external_ids: A list of all external ids created and referenced from the data model objects.
        """

        data_model_objects: dict[str, v1_data_classes.DomainModelWrite] = {}
        type_nodes = {}

        for folder in self.folders_to_process:
            all_data_model_files = list((folder).glob("*.yaml"))

            data_model_configuration_file = folder / "data_model_configuration.yaml"
            if data_model_configuration_file in all_data_model_files:
                all_data_model_files.remove(data_model_configuration_file)
                self.data_model_configuration = DataModelConfiguration.from_yaml(
                    self.data_model_classes, data_model_configuration_file
                )
            else:
                self.data_model_configuration = DataModelConfiguration.from_yaml(self.data_model_classes)
                logger.info(f"No data model configuration file found in {folder}, using default configuration")

            for data_model_file in all_data_model_files:
                # TODO: allow using prefix of file name to match type and add extra context to rest of file name?
                if data_model_file.stem not in self.data_model_configuration:
                    logger.warning(f"Skipping {data_model_file} as it is not in the supported types")
                    continue

                type_configuration = self.data_model_configuration[data_model_file.stem]

                raw_data_model_objects = load_yaml(data_model_file, "list")

                logger.info(f"Processing {data_model_file} as type {type_configuration.name}")

                objects_to_add: list[v1_data_classes.DomainModelWrite] = []
                for unprocessed_object in raw_data_model_objects:
                    results = self._dict_to_type_object(unprocessed_object, type_configuration)
                    objects_to_add.extend(results[0].values())
                    data_model_objects.update(results[0])

                grouped_objects_to_add: dict[type, list[v1_data_classes.DomainModelWrite]] = {}
                for obj in objects_to_add:
                    obj_type = type(obj)  # Get the type of the object
                    if obj_type not in grouped_objects_to_add:
                        grouped_objects_to_add[obj_type] = []  # Create a new list for this type
                    grouped_objects_to_add[obj_type].append(obj)  # Add the object to the list

                for obj_type, objects in grouped_objects_to_add.items():
                    node_type = self._type_objects_to_file(
                        objects,
                        folder.stem,
                        obj_type,
                        client,
                    )
                    type_nodes.update(node_type)

        if type_nodes:
            node_type_yaml_file = self.toolkit_directory / "types.node.yaml"

            sorted_data = sorted(list(type_nodes.values()), key=lambda x: x["externalId"])

            with Path(node_type_yaml_file).open("w") as file:
                yaml.dump(sorted_data, file, sort_keys=True)

        return list(data_model_objects.values()), list(data_model_objects.keys())

    def _dict_to_type_object(
        self,
        raw_data: dict,
        type_configuration: DataModelConfiguration,
    ) -> tuple[dict[str, v1_data_classes.DomainModelWrite], list[str]]:
        """Converts a dictionary object into the specified data type.

        Provided the data input and type configuration the dictionary will be updated with properties to be overridden
        from other sources and then converted into the specified data type.

        Args:
            data: A dictionary of data to be converted into the specified data type.
            type_configuration: A DataModelConfiguration object detailing the configuration of the data type.

        Returns:
            data_model_objects: A dictionary of all data model objects created from the input data.
            reference_external_ids: A list of all external ids created and referenced from the data model objects.

        Raises:
            ValueError: If any required fields are missing or invalid for the specified data type.
        """

        logger.info(f"Processing {raw_data} as type {type_configuration.type_.__name__}")

        # check if any fields from input data don't match the domain model
        check_input_keys(raw_data, list(type_configuration.properties.keys()))

        # split data into immutable, property configuration and reference data
        immutable_data, property_configuration_data, reference_data = self._split_data_by_parsing_method(raw_data)

        # populate properties from configurations
        parsed_properties, generated_data_model_objects = self._populate_properties_from_configurations(
            property_configuration_data, type_configuration.property_configurations, raw_data
        )

        # populate all external id references
        parsed_reference_data, reference_external_ids = parse_external_ids(
            reference_data, type_configuration.properties, self.data_model_classes
        )

        # join all dictionaries together in priority order
        parsed_data = self._join_data_by_parsing_method(immutable_data, parsed_properties, parsed_reference_data)

        # TODO: handle these special cases in a better way
        if type_configuration.type_ is v1_data_classes.PlantInformationWrite:
            parsed_data["generators"] = (
                self._get_all_connections(parsed_data, "plant", "generator")
                if "generators" not in parsed_data
                else parsed_data["generators"]
            )
            parsed_data["connection_losses"] = (
                self._get_connection_losses(parsed_data)
                if "connection_losses" not in parsed_data
                else parsed_data["connection_losses"]
            )

        # TODO: extract all sub objects into the data_model_objects list
        try:
            fdm_object = type_configuration.type_(**parsed_data)
            generated_data_model_objects[fdm_object.external_id] = fdm_object
        except ValidationError as exc:
            raise ValueError(
                f"Missing/invalid field for {type_configuration.type_.__name__} in data {parsed_data}"
            ) from exc

        return generated_data_model_objects, reference_external_ids

    def _type_objects_to_file(
        self,
        data_model_objects: list[v1_data_classes.DomainModelWrite],
        prefix_name: str,
        object_type: Any,
        client: CogniteClient,
    ) -> dict[str, dict[str, str]]:
        """Converts a data model object into a dictionary object matching API spec and writes it to a file.

        Provided the data model object input the object will be converted into a dictionary and written to a file.

        Args:
            data_model_object: A DomainModelWrite object to be converted into a dictionary and written to a file.
        """

        node_file_data = []
        node_external_ids = []
        edge_file_data = []
        edge_external_ids = []

        type_external_id = object_type.__name__.replace("Write", "")

        view = client.data_modeling.views.retrieve(ids=[("power_ops_core", type_external_id)])[0]
        properties = view.properties
        edge_properties = {}
        direct_relation_list_properties = {}
        direct_relation_properties = {}
        float_list_properties = {}
        float_properties = {}
        other_properties = {}

        for name, info in properties.items():
            if isinstance(info, MultiEdgeConnection):
                edge_properties[name] = info.type
            elif isinstance(info, MappedProperty):
                if isinstance(info.type, DirectRelation):
                    if info.type.is_list:
                        direct_relation_list_properties[name] = info.type
                    else:
                        direct_relation_properties[name] = info.type
                elif isinstance(info.type, Float64 | Float32):
                    if info.type.is_list:
                        float_list_properties[name] = info.type
                    else:
                        float_properties[name] = info.type
                elif isinstance(info.type, Text | TimeSeriesReference | Boolean | Json | FileReference | Int64 | Int32):
                    other_properties[name] = info.type
                else:
                    raise ValueError(f"Type mapped {info.type} is not supported")
            else:
                raise ValueError(f"Type {info} is not supported")

        for obj in data_model_objects:
            object_dump = obj.dump()
            object_dump.pop("data_record")
            object_dump.pop("space")
            object_dump.pop("node_type")

            external_id = object_dump.pop("externalId")

            edges = {}
            for property_name, info in object_dump.items():
                if property_name in edge_properties:
                    edges[property_name] = info
                elif property_name in direct_relation_properties and object_dump[property_name] is not None:
                    object_dump[property_name] = {
                        "space": self.instance_space,
                        "externalId": object_dump[property_name],
                    }
                elif property_name in direct_relation_list_properties:
                    object_dump[property_name] = [
                        {
                            "space": self.instance_space,
                            "externalId": item,
                        }
                        for item in object_dump[property_name]
                        if item is not None
                    ]
                elif property_name in float_properties:
                    if object_dump[property_name] is not None and object_dump[property_name].is_integer():
                        object_dump[property_name] = int(object_dump[property_name])
                elif property_name in float_list_properties:
                    object_dump[property_name] = [
                        int(num) if num.is_integer() else num for num in object_dump[property_name]
                    ]

            for edge_name in edges.keys():
                object_dump.pop(edge_name)

            for edge_name, edge_info in edges.items():
                if not isinstance(edge_info, list):
                    edge_info = [edge_info]  # type: ignore[assignment]
                for edge_end_external_id in edge_info:  # type: ignore[attr-defined]
                    if edge_end_external_id:
                        edge_object = {
                            "instanceType": "edge",
                            "type": {
                                "space": edge_properties[edge_name].space,
                                "externalId": edge_properties[edge_name].external_id,
                            },
                            "space": self.instance_space,
                            "externalId": f"{external_id}:{edge_end_external_id}",
                            "startNode": {
                                "space": self.instance_space,
                                "externalId": external_id,
                            },
                            "endNode": {
                                "space": self.instance_space,
                                "externalId": edge_end_external_id,
                            },
                        }

                        if edge_object["externalId"] not in edge_external_ids:
                            edge_external_ids.append(edge_object["externalId"])
                            edge_file_data.append(edge_object)
                        else:
                            logger.warning(f"Edge {edge_object['externalId']} already exists")

            if self.ignore_nones:
                object_dump = {k: v for k, v in object_dump.items() if v is not None}

            node_object = {
                "space": self.instance_space,
                "externalId": external_id,
                "instanceType": "node",
                "type": {
                    "space": self.type_space,
                    "externalId": type_external_id,
                },
                "sources": [
                    {
                        "source": {
                            "space": "power_ops_core",
                            "externalId": type_external_id,
                            "version": "1",
                            "type": "view",
                        },
                        "properties": object_dump,
                    }
                ],
            }
            if node_object["externalId"] not in node_external_ids:
                node_external_ids.append(node_object["externalId"])
                node_file_data.append(node_object)
            else:
                logger.warning(f"Node {node_object['externalId']} already exists")

        file_name_prefix = f"{prefix_name}:{type_external_id}"

        if node_file_data:
            node_yaml_file = self.toolkit_directory / f"{file_name_prefix}.node.yaml"

            sorted_data = sorted(node_file_data, key=lambda x: x["externalId"])

            with Path(node_yaml_file).open("w") as file:
                yaml.dump(sorted_data, file)

        if edge_file_data:
            edge_yaml_file = self.toolkit_directory / f"{file_name_prefix}.edge.yaml"

            sorted_data = sorted(edge_file_data, key=lambda x: x["externalId"])

            with Path(edge_yaml_file).open("w") as file:
                yaml.dump(sorted_data, file)

        type_node = {
            type_external_id: {
                "externalId": type_external_id,
                "space": self.type_space,
            }
        }
        return type_node

    @staticmethod
    def _split_data_by_parsing_method(
        raw_data: dict[str, Any],
    ) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
        """Splits an unparsed dictionary into a dictionary for each type of parsing should be applied.

        Provided the data input the dictionary will be split into immutable data, property configuration data and
        reference data. Where immutable data is the data that will not be overridden, property configuration data
        will be overridden from other sources based on the property configuration provided and reference data will
        be used to fetch the external ids.

        Args:
            raw_data: A dictionary of data to be split into the different parsing methods based on the values.

        Returns:
            immutable_data: A dictionary of data that will not be overridden by any other settings.
            property_configuration_data: A dictionary of data that will be overridden from other sources.
            reference_data: A dictionary of data that will be used to fetch the external ids.
        """

        immutable_data: dict[str, Any] = {}
        property_configuration_data: dict[str, Any] = {}
        reference_data: dict[str, Any] = {}

        for key, value in raw_data.items():
            if isinstance(value, str) and value.startswith("[SOURCE"):
                property_configuration_data[key] = value
            elif isinstance(value, str) and (value.startswith("[external_id") or value.startswith("[name")):
                reference_data[key] = value
            elif isinstance(value, list):
                immutable_subdata = []
                reference_subdata = []
                for item in value:
                    if isinstance(item, str) and (item.startswith("[external_id") or item.startswith("[name")):
                        reference_subdata.append(item)
                    else:
                        immutable_subdata.append(item)
                if reference_subdata:
                    reference_data[key] = reference_subdata
                if immutable_subdata:
                    immutable_data[key] = immutable_subdata
            else:
                immutable_data[key] = value

        return immutable_data, property_configuration_data, reference_data

    @staticmethod
    def _join_data_by_parsing_method(
        immutable_data: dict[str, Any], property_configuration_data: dict[str, Any], reference_data: dict[str, Any]
    ) -> dict[str, Any]:
        """Joins dictionaries together in a prioritized order based on the parsing method.

        Provided the data input the dictionaries will be joined together in a prioritized order based on the parsing
        method. The immutable data will override any other data, followed by the reference data and lastly the
        property configuration data so that any general property configurations will be overridden by the specific
        instance data.

        Args:
            immutable_data: A dictionary of data that will not be overridden by any other settings.
            property_configuration_data: A dictionary of data that will be overridden from other sources.
            reference_data: A dictionary of data that will be overridden or joined only with the immutable data.

        Returns:
            joined_data: A dictionary of all data joined together in a prioritized order based on the parsing method.
        """

        joined_data = property_configuration_data.copy()
        joined_data.update(reference_data)

        for key, value in immutable_data.items():
            if key in joined_data and isinstance(joined_data[key], list):
                joined_data[key].extend(value)
            else:
                joined_data[key] = value

        return joined_data

    def _populate_properties_from_configurations(
        self, property_configuration_data: dict, property_configurations: list[PropertyConfiguration], raw_data: dict
    ) -> tuple[dict[str, Any], dict[str, v1_data_classes.DomainModelWrite]]:
        """Populates properties from configurations based on the provided data.

        Provided the data input and property configurations the data will be parsed based on the property configurations
        and the data will be updated with the parsed properties and any generated data model objects. The property
        configurations will be prioritized based on the union of the default property configurations and the override
        property. The generated data model objects will be added to the generated_data_model_objects dictionary.

        Args:
            property_configuration_data: A dictionary of data that will be overridden from other sources.
            property_configurations: A list of PropertyConfiguration objects detailing the configuration of
                                     each property.
            raw_data: A dictionary of data to be referenced based on the property configurations.

        Returns:
            parsed_properties: A dictionary of all parsed properties based on the property configurations.
            generated_data_model_objects: A dictionary of all generated data model objects from any subtype properties.
        """

        override_property_configurations = PropertyConfiguration.from_data(property_configuration_data)
        all_properties_to_extract = PropertyConfiguration.union_lists(
            property_configurations, override_property_configurations
        )

        parsed_properties = {}
        generated_data_model_objects = {}
        for property_configuration in all_properties_to_extract:
            if property_configuration.is_subtype:
                if property_configuration.is_list:
                    # TODO: simplify list logic
                    subtype_list = self._get_subtype_list(property_configuration, raw_data)

                    parsed_properties[property_configuration.property] = [
                        subtype.external_id for subtype in subtype_list
                    ]
                    for subtype in subtype_list:
                        if subtype.external_id in generated_data_model_objects:
                            logger.warning(
                                f"Subtype {subtype.external_id} already exists in generated data model objects"
                            )
                        generated_data_model_objects[subtype.external_id] = subtype
                else:
                    subtype = self._get_subtype_from_dictionary(property_configuration, raw_data)

                    parsed_properties[property_configuration.property] = subtype.external_id
                    generated_data_model_objects[subtype.external_id] = subtype
            else:
                parsed_properties[property_configuration.property] = self._get_property_value_from_source(
                    property_configuration, raw_data
                )

        return parsed_properties, generated_data_model_objects

    # TODO: check if this is the correct way to handle connections, update to not use from_type/to_type?
    # https://github.com/cognitedata/power-ops-sdk/pull/347#discussion_r1594731007
    def _get_all_connections(self, data: dict, power_asset_type_a: str, power_asset_type_b: str) -> list[str]:
        """Creates a list of all connections based on the provided data and asset types.

        Provided the data input and asset types the connections will be filtered based on the asset types and the
        connections will be returned as a list of external ids.

        Args:
            data: A dictionary of data to be used to filter the connections.
            power_asset_type_a: A string of the first power asset type to filter the connections.
            power_asset_type_b: A string of the second power asset type to filter the connections.

        Returns:
            connections: A list of all connections external ids based on the provided data and asset types.
        """

        all_connections = []
        source_files = set()
        for dm_config in self.data_model_configuration.values():
            for property_config in dm_config.property_configurations:
                if property_config.source_file:
                    source_files.add(property_config.source_file)

        for source_file in source_files:
            all_connections += self._get_property_value_from_source(
                PropertyConfiguration(
                    property="connections",
                    source_file=source_file,
                    extraction_path="connections",
                    cast_type=None,
                ),
                data,
            )
        connections = []

        name = data["name"]

        for connection in all_connections:
            if (
                connection["from"] == name
                and connection.get("from_type") == power_asset_type_a
                and connection.get("to_type") == power_asset_type_b
            ):
                connections.append(connection["to"])
            elif (
                connection["to"] == name
                and connection.get("to_type") == power_asset_type_a
                and connection.get("from_type") == power_asset_type_b
            ):
                connections.append(connection["from"])

        # TODO: use external_id logic?
        return [f"{power_asset_type_b}_{asset}".lower() for asset in set(connections)]

    def _get_connection_losses(self, data: dict) -> float:
        """Gets the connection losses based on the provided data.

        TODO: Implement connection losses calculation based on the provided data.

        Provided the data input the connection losses will be calculated based on the provided data.

        Args:
            data: A dictionary of data to be used to filter the connections.

        Returns:
            connection_losses: A float of the connection losses based on the provided data.
        """

        return 0.0

    def _get_property_value_from_source(
        self,
        property_configuration: PropertyConfiguration,
        instance_data: Optional[dict] = None,
        source_data: Optional[dict] = None,
    ) -> Any:
        """Gets the property value based on the provided data and property configuration.

        TODO: Implement type casting and list handling, simplify logic.

        Provided the data input and property configuration the property value will be extracted based on the
        property configuration extraction_path and source_file. The property value will be cast to the specified type
        if provided.

        Args:
            property_configuration: A PropertyConfiguration object detailing the configuration of the property.
            instance_data: A dictionary of instance data to be used for any reference data based on the
                            property configuration.
            source_data: A dictionary of source data to be used to extract the property value. If not provided the
                         source data will be loaded from the source_file in the property configuration.

        Returns:
            property_value: The property value extracted based on the provided data and property configuration.

        Raises:
            ValueError: No source data provided for property configuration
            ValueError: Source data is not a dictionary
            ValueError: Source data is not indexable
        """

        if source_data is None:
            source_data = {}
        if instance_data is None:
            instance_data = {}
        if not source_data:
            if property_configuration.source_file:
                path = Path(property_configuration.source_file)
                source_data = load_yaml(path, "dict")
            else:
                if property_configuration.default_value:
                    return property_configuration.default_value
                raise ValueError("No source data or default value provided for property configuration")

        try:
            for key in property_configuration.extraction_path or []:
                match = re.match(r"\[(\w+)\]", key)
                if key == "value":
                    # TODO: handle value and key to not use next(iter)??
                    if isinstance(source_data, dict):
                        source_data = next(iter(source_data.values()))
                    else:
                        raise ValueError("Source data is not a dictionary")
                elif match:
                    if isinstance(source_data, (Sequence | Mapping)):
                        if match.group(1).isdigit():
                            source_data = source_data[int(match.group(1))]
                        else:
                            source_data = source_data[instance_data[match.group(1)]]
                    else:
                        raise ValueError("Source data is not indexable")
                elif isinstance(source_data, (Sequence | Mapping)):
                    source_data = source_data[key]
                else:
                    raise ValueError("Source data is not indexable")
        except KeyError:
            logger.warning(f"Key {key} not found in {property_configuration.source_file}, using default value")
            return property_configuration.default_value

        if property_configuration.is_list and property_configuration.cast_type:
            # TODO: handle for list of subtypes??
            return source_data
        elif property_configuration.is_list:
            return source_data

        if property_configuration.cast_type and source_data:
            return self.cast_into_type(source_data, property_configuration.cast_type)
        else:
            return source_data

    def _get_subtype_from_dictionary(
        self, property_configuration: PropertyConfiguration, instance_data: Optional[dict] = None
    ) -> Any:
        """Gets the subtype property value based on the provided data and property configuration.

        TODO: Implement subtype handling, simplify logic.

        Provided the data input and property configuration the property value will be extracted based on the
        property configuration extraction_path and source_file. The property value will be cast to the specified
        type if provided.

        Args:
            property_configuration: A PropertyConfiguration object detailing the configuration of the property.
            instance_data: A dictionary of instance data to be used for any reference data based on the property
                            configuration.

        Returns:
            property_value: The property value extracted based on the provided data and property configuration.

        Raises:
            ValueError: Type is not supported, add import to type
            ValueError: Property configuration requires a cast_type
        """

        if instance_data is None:
            instance_data = {}
        if property_configuration.cast_type not in self.data_model_configuration:
            raise ValueError(f"Type {property_configuration.cast_type} is not supported, add import to type")
        subtype_data_model_configuration = self.data_model_configuration[property_configuration.cast_type]

        subtype_data = {}
        for subtype_property_configuration in subtype_data_model_configuration.property_configurations:
            if not subtype_property_configuration.is_subtype:
                subtype_data[subtype_property_configuration.property] = self._get_property_value_from_source(
                    subtype_property_configuration, instance_data
                )
            else:
                # TODO: handle this scenario???
                logger.error("can't process")

        if property_configuration.cast_type:
            return self.cast_into_type(subtype_data, property_configuration.cast_type)
        else:
            raise ValueError(f"Property configuration {property_configuration} requires a cast_type")

    def _get_subtype_list(
        self, property_configuration: PropertyConfiguration, instance_data: Optional[dict] = None
    ) -> list[Any]:
        """Gets a list of subtype property values based on the provided data and property configuration.

        TODO: Implement subtype handling, simplify logic.

        Provided the data input and property configuration the property value will be extracted based on the property
        configuration extraction_path and source_file. The property value will be cast to the specified
        type if provided.

        Args:
            property_configuration: A PropertyConfiguration object detailing the configuration of the property.
            instance_data: A dictionary of instance data to be used for any reference data based on the property
                            configuration.

        Returns:
            parsed_list_data: A list of property values extracted based on the provided data and property configuration.

        Raises:
            ValueError: Type is not supported, add import to type
            ValueError: Property configuration requires a cast_type
        """

        if instance_data is None:
            instance_data = {}
        list_data = self._get_property_value_from_source(property_configuration, instance_data)
        if not list_data:
            return []
        if property_configuration.cast_type not in self.data_model_configuration:
            raise ValueError(f"Type {property_configuration.cast_type} is not supported, add import to type")
        subtype_data_model_configuration = self.data_model_configuration[property_configuration.cast_type]

        parsed_list_data = []
        for source_data in list_data:
            subtype_data = {}
            for subtype_property_configuration in subtype_data_model_configuration.property_configurations:
                if not subtype_property_configuration.is_subtype:
                    subtype_data[subtype_property_configuration.property] = self._get_property_value_from_source(
                        subtype_property_configuration, instance_data, source_data
                    )  # TODO: get value from source
                else:
                    # TODO: handle this scenario???
                    pass
            if property_configuration.cast_type:
                parsed_list_data.append(self.cast_into_type(subtype_data, property_configuration.cast_type))
            else:
                raise ValueError(f"Property configuration {property_configuration} requires a cast_type")

        return parsed_list_data

    def cast_into_type(self, data: dict[str, Any], cast_type: str) -> type:
        """Casts a dictionary into a type.

        Provided the data input and type the dictionary will be cast into the specified type.

        Args:
            data: A dictionary of data to be cast into the specified type.
            type_: A DomainModelWrite object detailing the type to cast the data into.

        Returns:
            fdm_object: A DomainModelWrite object cast from the data input.
        """
        primitive_type_mapping = {
            "int": int,
            "float": float,
            "str": str,
        }

        if cast_type in primitive_type_mapping:
            return primitive_type_mapping[cast_type](data)

        if cast_type not in self.data_model_classes:
            raise ValueError(f"Type {cast_type} is not supported")
        type_ = self.data_model_classes[cast_type]

        return type_(**data)



import warnings
from typing import TypeVar

T = TypeVar("T")


def deprecated_class(cls: type[T]) -> type[T]:
    """Decorator to mark a class as deprecated, showing a message only once."""

    def new_getattribute(self, name):  # type: ignore[no-untyped-def]
        warnings.warn(
            f"\nWarning: Class `{cls.__name__}` is deprecated and will be removed.",
            DeprecationWarning,
            stacklevel=2,
        )
        return super(cls, self).__getattribute__(name)

    cls.__getattribute__ = new_getattribute  # type: ignore[method-assign]

    return cls



from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from cognite.client.data_classes import (
    TimeSeries as CogniteTimeSeries,
    TimeSeriesWrite as CogniteTimeSeriesWrite,
    Sequence as CogniteSequence,
    SequenceWrite as CogniteSequenceWrite,
)
from pydantic import Field
from pydantic import field_validator, model_validator, ValidationInfo

from cognite.powerops.client._generated.v1.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    FileMetadata,
    FileMetadataWrite,
    FileMetadataGraphQL,
    TimeSeries,
    TimeSeriesWrite,
    TimeSeriesGraphQL,
    TimeSeriesReferenceAPI,
    SequenceRead,
    SequenceWrite,
    SequenceGraphQL,
    T_DomainModelList,
    as_node_id,
    as_read_args,
    as_write_args,
    is_tuple_id,
    as_instance_dict_id,
    parse_single_connection,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    ViewPropertyId,
    DirectRelationFilter,
    FloatFilter,
)
from cognite.powerops.client._generated.v1.data_classes._partial_bid_matrix_information import PartialBidMatrixInformation, PartialBidMatrixInformationWrite
if TYPE_CHECKING:
    from cognite.powerops.client._generated.v1.data_classes._alert import Alert, AlertList, AlertGraphQL, AlertWrite, AlertWriteList
    from cognite.powerops.client._generated.v1.data_classes._bid_matrix import BidMatrix, BidMatrixList, BidMatrixGraphQL, BidMatrixWrite, BidMatrixWriteList
    from cognite.powerops.client._generated.v1.data_classes._partial_bid_configuration import PartialBidConfiguration, PartialBidConfigurationList, PartialBidConfigurationGraphQL, PartialBidConfigurationWrite, PartialBidConfigurationWriteList
    from cognite.powerops.client._generated.v1.data_classes._power_asset import PowerAsset, PowerAssetList, PowerAssetGraphQL, PowerAssetWrite, PowerAssetWriteList
    from cognite.powerops.client._generated.v1.data_classes._price_production import PriceProduction, PriceProductionList, PriceProductionGraphQL, PriceProductionWrite, PriceProductionWriteList


__all__ = [
    "PartialBidMatrixInformationWithScenarios",
    "PartialBidMatrixInformationWithScenariosWrite",
    "PartialBidMatrixInformationWithScenariosApply",
    "PartialBidMatrixInformationWithScenariosList",
    "PartialBidMatrixInformationWithScenariosWriteList",
    "PartialBidMatrixInformationWithScenariosApplyList",
    "PartialBidMatrixInformationWithScenariosFields",
    "PartialBidMatrixInformationWithScenariosTextFields",
    "PartialBidMatrixInformationWithScenariosGraphQL",
]


PartialBidMatrixInformationWithScenariosTextFields = Literal["external_id", "state", "bid_matrix", "linked_time_series"]
PartialBidMatrixInformationWithScenariosFields = Literal["external_id", "state", "bid_matrix", "linked_time_series", "resource_cost"]

_PARTIALBIDMATRIXINFORMATIONWITHSCENARIOS_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "state": "state",
    "bid_matrix": "bidMatrix",
    "linked_time_series": "linkedTimeSeries",
    "resource_cost": "resourceCost",
}


class PartialBidMatrixInformationWithScenariosGraphQL(GraphQLCore):
    """This represents the reading version of partial bid matrix information with scenario, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid matrix information with scenario.
        data_record: The data record of the partial bid matrix information with scenario node.
        state: The state field.
        bid_matrix: The bid matrix field.
        linked_time_series: The linked time series field.
        alerts: An array of calculation level Alerts.
        underlying_bid_matrices: An array of intermediate BidMatrices.
        power_asset: The power asset field.
        resource_cost: The resource cost field.
        partial_bid_configuration: The partial bid configuration field.
        multi_scenario_input: TODO
    """

    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PartialBidMatrixInformationWithScenarios", "1")
    state: Optional[str] = None
    bid_matrix: Optional[SequenceGraphQL] = Field(None, alias="bidMatrix")
    linked_time_series: Optional[list[TimeSeriesGraphQL]] = Field(None, alias="linkedTimeSeries")
    alerts: Optional[list[AlertGraphQL]] = Field(default=None, repr=False)
    underlying_bid_matrices: Optional[list[BidMatrixGraphQL]] = Field(default=None, repr=False, alias="underlyingBidMatrices")
    power_asset: Optional[PowerAssetGraphQL] = Field(default=None, repr=False, alias="powerAsset")
    resource_cost: Optional[float] = Field(None, alias="resourceCost")
    partial_bid_configuration: Optional[PartialBidConfigurationGraphQL] = Field(default=None, repr=False, alias="partialBidConfiguration")
    multi_scenario_input: Optional[list[PriceProductionGraphQL]] = Field(default=None, repr=False, alias="multiScenarioInput")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("linked_time_series", mode="before")
    def clean_list(cls, value: Any) -> Any:
        if isinstance(value, list):
            return [v for v in value if v is not None] or None
        return value

    @field_validator("alerts", "underlying_bid_matrices", "power_asset", "partial_bid_configuration", "multi_scenario_input", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> PartialBidMatrixInformationWithScenarios:
        """Convert this GraphQL format of partial bid matrix information with scenario to the reading format."""
        return PartialBidMatrixInformationWithScenarios.model_validate(as_read_args(self))

    def as_write(self) -> PartialBidMatrixInformationWithScenariosWrite:
        """Convert this GraphQL format of partial bid matrix information with scenario to the writing format."""
        return PartialBidMatrixInformationWithScenariosWrite.model_validate(as_write_args(self))


class PartialBidMatrixInformationWithScenarios(PartialBidMatrixInformation):
    """This represents the reading version of partial bid matrix information with scenario.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid matrix information with scenario.
        data_record: The data record of the partial bid matrix information with scenario node.
        state: The state field.
        bid_matrix: The bid matrix field.
        linked_time_series: The linked time series field.
        alerts: An array of calculation level Alerts.
        underlying_bid_matrices: An array of intermediate BidMatrices.
        power_asset: The power asset field.
        resource_cost: The resource cost field.
        partial_bid_configuration: The partial bid configuration field.
        multi_scenario_input: TODO
    """

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PartialBidMatrixInformationWithScenarios", "1")

    node_type: Union[dm.DirectRelationReference, None] = None
    multi_scenario_input: Optional[list[Union[PriceProduction, str, dm.NodeId]]] = Field(default=None, repr=False, alias="multiScenarioInput")
    @field_validator("power_asset", "partial_bid_configuration", mode="before")
    @classmethod
    def parse_single(cls, value: Any, info: ValidationInfo) -> Any:
        return parse_single_connection(value, info.field_name)

    @field_validator("alerts", "underlying_bid_matrices", "multi_scenario_input", mode="before")
    @classmethod
    def parse_list(cls, value: Any, info: ValidationInfo) -> Any:
        if value is None:
            return None
        return [parse_single_connection(item, info.field_name) for item in value]

    def as_write(self) -> PartialBidMatrixInformationWithScenariosWrite:
        """Convert this read version of partial bid matrix information with scenario to the writing version."""
        return PartialBidMatrixInformationWithScenariosWrite.model_validate(as_write_args(self))

    def as_apply(self) -> PartialBidMatrixInformationWithScenariosWrite:
        """Convert this read version of partial bid matrix information with scenario to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PartialBidMatrixInformationWithScenariosWrite(PartialBidMatrixInformationWrite):
    """This represents the writing version of partial bid matrix information with scenario.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid matrix information with scenario.
        data_record: The data record of the partial bid matrix information with scenario node.
        state: The state field.
        bid_matrix: The bid matrix field.
        linked_time_series: The linked time series field.
        alerts: An array of calculation level Alerts.
        underlying_bid_matrices: An array of intermediate BidMatrices.
        power_asset: The power asset field.
        resource_cost: The resource cost field.
        partial_bid_configuration: The partial bid configuration field.
        multi_scenario_input: TODO
    """
    _container_fields: ClassVar[tuple[str, ...]] = ("bid_matrix", "linked_time_series", "partial_bid_configuration", "power_asset", "resource_cost", "state",)
    _outwards_edges: ClassVar[tuple[tuple[str, dm.DirectRelationReference], ...]] = (("alerts", dm.DirectRelationReference("power_ops_types", "calculationIssue")), ("multi_scenario_input", dm.DirectRelationReference("power_ops_types", "calculationIssue")), ("underlying_bid_matrices", dm.DirectRelationReference("power_ops_types", "intermediateBidMatrix")),)
    _direct_relations: ClassVar[tuple[str, ...]] = ("partial_bid_configuration", "power_asset",)

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PartialBidMatrixInformationWithScenarios", "1")

    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = None
    multi_scenario_input: Optional[list[Union[PriceProductionWrite, str, dm.NodeId]]] = Field(default=None, repr=False, alias="multiScenarioInput")

    @field_validator("multi_scenario_input", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value


class PartialBidMatrixInformationWithScenariosApply(PartialBidMatrixInformationWithScenariosWrite):
    def __new__(cls, *args, **kwargs) -> PartialBidMatrixInformationWithScenariosApply:
        warnings.warn(
            "PartialBidMatrixInformationWithScenariosApply is deprecated and will be removed in v1.0. "
            "Use PartialBidMatrixInformationWithScenariosWrite instead. "
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PartialBidMatrixInformationWithScenarios.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)

class PartialBidMatrixInformationWithScenariosList(DomainModelList[PartialBidMatrixInformationWithScenarios]):
    """List of partial bid matrix information with scenarios in the read version."""

    _INSTANCE = PartialBidMatrixInformationWithScenarios
    def as_write(self) -> PartialBidMatrixInformationWithScenariosWriteList:
        """Convert these read versions of partial bid matrix information with scenario to the writing versions."""
        return PartialBidMatrixInformationWithScenariosWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PartialBidMatrixInformationWithScenariosWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()

    @property
    def alerts(self) -> AlertList:
        from ._alert import Alert, AlertList
        return AlertList([item for items in self.data for item in items.alerts or [] if isinstance(item, Alert)])

    @property
    def underlying_bid_matrices(self) -> BidMatrixList:
        from ._bid_matrix import BidMatrix, BidMatrixList
        return BidMatrixList([item for items in self.data for item in items.underlying_bid_matrices or [] if isinstance(item, BidMatrix)])

    @property
    def power_asset(self) -> PowerAssetList:
        from ._power_asset import PowerAsset, PowerAssetList
        return PowerAssetList([item.power_asset for item in self.data if isinstance(item.power_asset, PowerAsset)])
    @property
    def partial_bid_configuration(self) -> PartialBidConfigurationList:
        from ._partial_bid_configuration import PartialBidConfiguration, PartialBidConfigurationList
        return PartialBidConfigurationList([item.partial_bid_configuration for item in self.data if isinstance(item.partial_bid_configuration, PartialBidConfiguration)])
    @property
    def multi_scenario_input(self) -> PriceProductionList:
        from ._price_production import PriceProduction, PriceProductionList
        return PriceProductionList([item for items in self.data for item in items.multi_scenario_input or [] if isinstance(item, PriceProduction)])


class PartialBidMatrixInformationWithScenariosWriteList(DomainModelWriteList[PartialBidMatrixInformationWithScenariosWrite]):
    """List of partial bid matrix information with scenarios in the writing version."""

    _INSTANCE = PartialBidMatrixInformationWithScenariosWrite
    @property
    def alerts(self) -> AlertWriteList:
        from ._alert import AlertWrite, AlertWriteList
        return AlertWriteList([item for items in self.data for item in items.alerts or [] if isinstance(item, AlertWrite)])

    @property
    def underlying_bid_matrices(self) -> BidMatrixWriteList:
        from ._bid_matrix import BidMatrixWrite, BidMatrixWriteList
        return BidMatrixWriteList([item for items in self.data for item in items.underlying_bid_matrices or [] if isinstance(item, BidMatrixWrite)])

    @property
    def power_asset(self) -> PowerAssetWriteList:
        from ._power_asset import PowerAssetWrite, PowerAssetWriteList
        return PowerAssetWriteList([item.power_asset for item in self.data if isinstance(item.power_asset, PowerAssetWrite)])
    @property
    def partial_bid_configuration(self) -> PartialBidConfigurationWriteList:
        from ._partial_bid_configuration import PartialBidConfigurationWrite, PartialBidConfigurationWriteList
        return PartialBidConfigurationWriteList([item.partial_bid_configuration for item in self.data if isinstance(item.partial_bid_configuration, PartialBidConfigurationWrite)])
    @property
    def multi_scenario_input(self) -> PriceProductionWriteList:
        from ._price_production import PriceProductionWrite, PriceProductionWriteList
        return PriceProductionWriteList([item for items in self.data for item in items.multi_scenario_input or [] if isinstance(item, PriceProductionWrite)])


class PartialBidMatrixInformationWithScenariosApplyList(PartialBidMatrixInformationWithScenariosWriteList): ...


def _create_partial_bid_matrix_information_with_scenario_filter(
    view_id: dm.ViewId,
    state: str | list[str] | None = None,
    state_prefix: str | None = None,
    power_asset: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    min_resource_cost: float | None = None,
    max_resource_cost: float | None = None,
    partial_bid_configuration: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(state, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("state"), value=state))
    if state and isinstance(state, list):
        filters.append(dm.filters.In(view_id.as_property_ref("state"), values=state))
    if state_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("state"), value=state_prefix))
    if isinstance(power_asset, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(power_asset):
        filters.append(dm.filters.Equals(view_id.as_property_ref("powerAsset"), value=as_instance_dict_id(power_asset)))
    if power_asset and isinstance(power_asset, Sequence) and not isinstance(power_asset, str) and not is_tuple_id(power_asset):
        filters.append(dm.filters.In(view_id.as_property_ref("powerAsset"), values=[as_instance_dict_id(item) for item in power_asset]))
    if min_resource_cost is not None or max_resource_cost is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("resourceCost"), gte=min_resource_cost, lte=max_resource_cost))
    if isinstance(partial_bid_configuration, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(partial_bid_configuration):
        filters.append(dm.filters.Equals(view_id.as_property_ref("partialBidConfiguration"), value=as_instance_dict_id(partial_bid_configuration)))
    if partial_bid_configuration and isinstance(partial_bid_configuration, Sequence) and not isinstance(partial_bid_configuration, str) and not is_tuple_id(partial_bid_configuration):
        filters.append(dm.filters.In(view_id.as_property_ref("partialBidConfiguration"), values=[as_instance_dict_id(item) for item in partial_bid_configuration]))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _PartialBidMatrixInformationWithScenariosQuery(NodeQueryCore[T_DomainModelList, PartialBidMatrixInformationWithScenariosList]):
    _view_id = PartialBidMatrixInformationWithScenarios._view_id
    _result_cls = PartialBidMatrixInformationWithScenarios
    _result_list_cls_end = PartialBidMatrixInformationWithScenariosList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.ResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_property: ViewPropertyId | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.ResultSetExpression | None = None,
    ):
        from ._alert import _AlertQuery
        from ._bid_matrix import _BidMatrixQuery
        from ._partial_bid_configuration import _PartialBidConfigurationQuery
        from ._power_asset import _PowerAssetQuery
        from ._price_production import _PriceProductionQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_property,
            connection_type,
            reverse_expression,
        )

        if _AlertQuery not in created_types:
            self.alerts = _AlertQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="alerts",
                connection_property=ViewPropertyId(self._view_id, "alerts"),
            )

        if _BidMatrixQuery not in created_types:
            self.underlying_bid_matrices = _BidMatrixQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="underlying_bid_matrices",
                connection_property=ViewPropertyId(self._view_id, "underlyingBidMatrices"),
            )

        if _PowerAssetQuery not in created_types:
            self.power_asset = _PowerAssetQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("powerAsset"),
                    direction="outwards",
                ),
                connection_name="power_asset",
                connection_property=ViewPropertyId(self._view_id, "powerAsset"),
            )

        if _PartialBidConfigurationQuery not in created_types:
            self.partial_bid_configuration = _PartialBidConfigurationQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("partialBidConfiguration"),
                    direction="outwards",
                ),
                connection_name="partial_bid_configuration",
                connection_property=ViewPropertyId(self._view_id, "partialBidConfiguration"),
            )

        if _PriceProductionQuery not in created_types:
            self.multi_scenario_input = _PriceProductionQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.EdgeResultSetExpression(
                    direction="outwards",
                    chain_to="destination",
                ),
                connection_name="multi_scenario_input",
                connection_property=ViewPropertyId(self._view_id, "multiScenarioInput"),
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.state = StringFilter(self, self._view_id.as_property_ref("state"))
        self.power_asset_filter = DirectRelationFilter(self, self._view_id.as_property_ref("powerAsset"))
        self.resource_cost = FloatFilter(self, self._view_id.as_property_ref("resourceCost"))
        self.partial_bid_configuration_filter = DirectRelationFilter(self, self._view_id.as_property_ref("partialBidConfiguration"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.state,
            self.power_asset_filter,
            self.resource_cost,
            self.partial_bid_configuration_filter,
        ])
        self.linked_time_series = TimeSeriesReferenceAPI(client,  lambda limit: [
            ts if isinstance(ts, str) else ts.external_id #type: ignore[misc]
            for item in self._list(limit=limit)
            if item.linked_time_series is not None
            for ts in item.linked_time_series
            if ts is not None and
               (isinstance(ts, str) or ts.external_id is not None)
        ])

    def list_partial_bid_matrix_information_with_scenario(self, limit: int = DEFAULT_QUERY_LIMIT) -> PartialBidMatrixInformationWithScenariosList:
        return self._list(limit=limit)


class PartialBidMatrixInformationWithScenariosQuery(_PartialBidMatrixInformationWithScenariosQuery[PartialBidMatrixInformationWithScenariosList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, PartialBidMatrixInformationWithScenariosList)



from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Optional, Union

from cognite.client import data_modeling as dm, CogniteClient
from cognite.client.data_classes import (
    TimeSeries as CogniteTimeSeries,
    TimeSeriesWrite as CogniteTimeSeriesWrite,
)
from pydantic import Field
from pydantic import field_validator, model_validator, ValidationInfo

from cognite.powerops.client._generated.v1.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelation,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    FileMetadata,
    FileMetadataWrite,
    FileMetadataGraphQL,
    TimeSeries,
    TimeSeriesWrite,
    TimeSeriesGraphQL,
    TimeSeriesReferenceAPI,
    T_DomainModelList,
    as_node_id,
    as_read_args,
    as_write_args,
    is_tuple_id,
    as_instance_dict_id,
    parse_single_connection,
    QueryCore,
    NodeQueryCore,
    StringFilter,
    ViewPropertyId,
    DirectRelationFilter,
)
if TYPE_CHECKING:
    from cognite.powerops.client._generated.v1.data_classes._shop_result import ShopResult, ShopResultList, ShopResultGraphQL, ShopResultWrite, ShopResultWriteList


__all__ = [
    "PriceProduction",
    "PriceProductionWrite",
    "PriceProductionApply",
    "PriceProductionList",
    "PriceProductionWriteList",
    "PriceProductionApplyList",
    "PriceProductionFields",
    "PriceProductionTextFields",
    "PriceProductionGraphQL",
]


PriceProductionTextFields = Literal["external_id", "name", "price", "production"]
PriceProductionFields = Literal["external_id", "name", "price", "production"]

_PRICEPRODUCTION_PROPERTIES_BY_FIELD = {
    "external_id": "externalId",
    "name": "name",
    "price": "price",
    "production": "production",
}


class PriceProductionGraphQL(GraphQLCore):
    """This represents the reading version of price production, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price production.
        data_record: The data record of the price production node.
        name: The name field.
        price: The price field.
        production: The production field.
        shop_result: The shop result field.
    """

    view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PriceProduction", "1")
    name: Optional[str] = None
    price: Optional[TimeSeriesGraphQL] = None
    production: Optional[TimeSeriesGraphQL] = None
    shop_result: Optional[ShopResultGraphQL] = Field(default=None, repr=False, alias="shopResult")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values


    @field_validator("shop_result", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> PriceProduction:
        """Convert this GraphQL format of price production to the reading format."""
        return PriceProduction.model_validate(as_read_args(self))

    def as_write(self) -> PriceProductionWrite:
        """Convert this GraphQL format of price production to the writing format."""
        return PriceProductionWrite.model_validate(as_write_args(self))


class PriceProduction(DomainModel):
    """This represents the reading version of price production.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price production.
        data_record: The data record of the price production node.
        name: The name field.
        price: The price field.
        production: The production field.
        shop_result: The shop result field.
    """

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PriceProduction", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("power_ops_types", "PriceProduction")
    name: str
    price: Union[TimeSeries, str, None] = None
    production: Union[TimeSeries, str, None] = None
    shop_result: Union[ShopResult, str, dm.NodeId, None] = Field(default=None, repr=False, alias="shopResult")
    @field_validator("shop_result", mode="before")
    @classmethod
    def parse_single(cls, value: Any, info: ValidationInfo) -> Any:
        return parse_single_connection(value, info.field_name)


    def as_write(self) -> PriceProductionWrite:
        """Convert this read version of price production to the writing version."""
        return PriceProductionWrite.model_validate(as_write_args(self))

    def as_apply(self) -> PriceProductionWrite:
        """Convert this read version of price production to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceProductionWrite(DomainModelWrite):
    """This represents the writing version of price production.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price production.
        data_record: The data record of the price production node.
        name: The name field.
        price: The price field.
        production: The production field.
        shop_result: The shop result field.
    """
    _container_fields: ClassVar[tuple[str, ...]] = ("name", "price", "production", "shop_result",)
    _direct_relations: ClassVar[tuple[str, ...]] = ("shop_result",)

    _view_id: ClassVar[dm.ViewId] = dm.ViewId("power_ops_core", "PriceProduction", "1")

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, dm.NodeId, tuple[str, str], None] = dm.DirectRelationReference("power_ops_types", "PriceProduction")
    name: str
    price: Union[TimeSeriesWrite, str, None] = None
    production: Union[TimeSeriesWrite, str, None] = None
    shop_result: Union[ShopResultWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="shopResult")

    @field_validator("shop_result", mode="before")
    def as_node_id(cls, value: Any) -> Any:
        if isinstance(value, dm.DirectRelationReference):
            return dm.NodeId(value.space, value.external_id)
        elif isinstance(value, tuple) and len(value) == 2 and all(isinstance(item, str) for item in value):
            return dm.NodeId(value[0], value[1])
        elif isinstance(value, list):
            return [cls.as_node_id(item) for item in value]
        return value


class PriceProductionApply(PriceProductionWrite):
    def __new__(cls, *args, **kwargs) -> PriceProductionApply:
        warnings.warn(
            "PriceProductionApply is deprecated and will be removed in v1.0. "
            "Use PriceProductionWrite instead. "
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PriceProduction.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)

class PriceProductionList(DomainModelList[PriceProduction]):
    """List of price productions in the read version."""

    _INSTANCE = PriceProduction
    def as_write(self) -> PriceProductionWriteList:
        """Convert these read versions of price production to the writing versions."""
        return PriceProductionWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PriceProductionWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()

    @property
    def shop_result(self) -> ShopResultList:
        from ._shop_result import ShopResult, ShopResultList
        return ShopResultList([item.shop_result for item in self.data if isinstance(item.shop_result, ShopResult)])

class PriceProductionWriteList(DomainModelWriteList[PriceProductionWrite]):
    """List of price productions in the writing version."""

    _INSTANCE = PriceProductionWrite
    @property
    def shop_result(self) -> ShopResultWriteList:
        from ._shop_result import ShopResultWrite, ShopResultWriteList
        return ShopResultWriteList([item.shop_result for item in self.data if isinstance(item.shop_result, ShopResultWrite)])

class PriceProductionApplyList(PriceProductionWriteList): ...


def _create_price_production_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    shop_result: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters: list[dm.Filter] = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(shop_result, str | dm.NodeId | dm.DirectRelationReference) or is_tuple_id(shop_result):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopResult"), value=as_instance_dict_id(shop_result)))
    if shop_result and isinstance(shop_result, Sequence) and not isinstance(shop_result, str) and not is_tuple_id(shop_result):
        filters.append(dm.filters.In(view_id.as_property_ref("shopResult"), values=[as_instance_dict_id(item) for item in shop_result]))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None


class _PriceProductionQuery(NodeQueryCore[T_DomainModelList, PriceProductionList]):
    _view_id = PriceProduction._view_id
    _result_cls = PriceProduction
    _result_list_cls_end = PriceProductionList

    def __init__(
        self,
        created_types: set[type],
        creation_path: list[QueryCore],
        client: CogniteClient,
        result_list_cls: type[T_DomainModelList],
        expression: dm.query.ResultSetExpression | None = None,
        connection_name: str | None = None,
        connection_property: ViewPropertyId | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        reverse_expression: dm.query.ResultSetExpression | None = None,
    ):
        from ._shop_result import _ShopResultQuery

        super().__init__(
            created_types,
            creation_path,
            client,
            result_list_cls,
            expression,
            dm.filters.HasData(views=[self._view_id]),
            connection_name,
            connection_property,
            connection_type,
            reverse_expression,
        )

        if _ShopResultQuery not in created_types:
            self.shop_result = _ShopResultQuery(
                created_types.copy(),
                self._creation_path,
                client,
                result_list_cls,
                dm.query.NodeResultSetExpression(
                    through=self._view_id.as_property_ref("shopResult"),
                    direction="outwards",
                ),
                connection_name="shop_result",
                connection_property=ViewPropertyId(self._view_id, "shopResult"),
            )

        self.space = StringFilter(self, ["node", "space"])
        self.external_id = StringFilter(self, ["node", "externalId"])
        self.name = StringFilter(self, self._view_id.as_property_ref("name"))
        self.shop_result_filter = DirectRelationFilter(self, self._view_id.as_property_ref("shopResult"))
        self._filter_classes.extend([
            self.space,
            self.external_id,
            self.name,
            self.shop_result_filter,
        ])
        self.price = TimeSeriesReferenceAPI(client,  lambda limit: [
            item.price if isinstance(item.price, str) else item.price.external_id #type: ignore[misc]
            for item in self._list(limit=limit)
            if item.price is not None and
               (isinstance(item.price, str) or item.price.external_id is not None)
        ])
        self.production = TimeSeriesReferenceAPI(client,  lambda limit: [
            item.production if isinstance(item.production, str) else item.production.external_id #type: ignore[misc]
            for item in self._list(limit=limit)
            if item.production is not None and
               (isinstance(item.production, str) or item.production.external_id is not None)
        ])

    def list_price_production(self, limit: int = DEFAULT_QUERY_LIMIT) -> PriceProductionList:
        return self._list(limit=limit)


class PriceProductionQuery(_PriceProductionQuery[PriceProductionList]):
    def __init__(self, client: CogniteClient):
        super().__init__(set(), [], client, PriceProductionList)



from __future__ import annotations

import logging
import os
import sys
from pathlib import Path
from string import Template
from typing import Literal

import yaml
from cognite.client import CogniteClient, global_config

from cognite.powerops.client.shop.cogshop_api import CogShopAPI
from cognite.pygen.utils.external_id_factories import ExternalIdFactory

from ._generated.v1 import PowerOpsModelsV1Client
from ._generated.v1.data_classes._core import DomainModelWrite
from .data_set_api import DataSetsAPI
from .shop.shop_run_api import SHOPRunAPI

# max_domain = max_total (255) - uuid (32) + separator (1)  noqa: ERA001
_MAX_DOMAIN_LENGTH = 233

logger = logging.getLogger(__name__)


class PowerOpsClient:
    def __init__(
        self,
        client: CogniteClient,
        read_dataset: str,
        write_dataset: str,
        monitor_dataset: str | None = None,
        cog_shop_service: Literal["prod", "staging"] | None = None,
    ):
        self.cdf = client
        self.datasets = DataSetsAPI(self.cdf, read_dataset, write_dataset, monitor_dataset)
        self.shop = SHOPRunAPI(self.cdf, self.datasets.write_dataset_id)
        self.v1 = PowerOpsModelsV1Client(self.cdf)
        self.cogshop = CogShopAPI(self.cdf, self.v1, cog_shop_service)

        DomainModelWrite.external_id_factory = ExternalIdFactory.create_external_id_factory(
            prefix_ext_id_factory=ExternalIdFactory(
                ExternalIdFactory.domain_name_factory(),
                shorten_length=_MAX_DOMAIN_LENGTH,
            ),
            override_external_id=False,
        )

    @classmethod
    def from_config(
        cls,
        config_path: Path | str,
    ) -> PowerOpsClient:
        """
        Create a PowerOpsClient from a configuration file.

        Args:
            config_path: The path to a yaml configuration file.

        Returns:
            A PowerOpsClient object.
        """

        if isinstance(config_path, str):
            config_path = Path(config_path)

        # Read in yaml file
        env_sub_template = Template(config_path.read_text())

        env_dict = dict(os.environ)

        if sys.version >= "3.11":  # Template.get_identifiers() is only available in Python 3.11 and later
            #  Fetch all environment variables referenced in the file to check if they are set
            all_identifiers = env_sub_template.get_identifiers()
            missing_env_vars = set(all_identifiers) - set(env_dict.keys())
            if missing_env_vars:
                raise ValueError(f"Missing environment variables: {missing_env_vars}")

        # Substitute environment variables in the file string
        file_env_parsed = env_sub_template.substitute(env_dict)

        # Load yaml file string into a dictionary to parse global and client configurations
        cognite_config = yaml.safe_load(file_env_parsed)

        # If you want to set a global configuration it must be done before creating the client
        if "global" in cognite_config:
            global_config.apply_settings(cognite_config["global"])

        client = CogniteClient.load(cognite_config["client"])

        power_ops_config = cognite_config.get("power_ops", {})

        return cls(client=client, **power_ops_config)



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import TYPE_CHECKING, cast

from cognite.client import data_modeling as dm, CogniteClient

from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    BenchmarkingCalculationInput,
)
from cognite.powerops.client._generated.v1.data_classes._core import (
    DEFAULT_QUERY_LIMIT,
    ViewPropertyId,
    T_DomainModel,
    T_DomainModelList,
    QueryBuilder,
    QueryStep,
)
from cognite.powerops.client._generated.v1.data_classes._shop_result import (
    _create_shop_result_filter,
)
from cognite.powerops.client._generated.v1._api._core import (
    QueryAPI,
    _create_edge_filter,
)

if TYPE_CHECKING:
    from cognite.powerops.client._generated.v1._api.shop_result_query import ShopResultQueryAPI


class BenchmarkingCalculationInputQueryAPI(QueryAPI[T_DomainModel, T_DomainModelList]):
    _view_id = dm.ViewId("power_ops_core", "BenchmarkingCalculationInput", "1")

    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder,
        result_cls: type[T_DomainModel],
        result_list_cls: type[T_DomainModelList],
        connection_property: ViewPropertyId | None = None,
        filter_: dm.filters.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ):
        super().__init__(client, builder, result_cls, result_list_cls)
        from_ = self._builder.get_from()
        self._builder.append(
            QueryStep(
                name=self._builder.create_name(from_),
                expression=dm.query.NodeResultSetExpression(
                    from_=from_,
                    filter=filter_,
                ),
                max_retrieve_limit=limit,
                view_id=self._view_id,
                connection_property=connection_property,
            )
        )
    def shop_results(
        self,
        case: str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference | Sequence[str | tuple[str, str] | dm.NodeId | dm.DirectRelationReference] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        external_id_prefix_edge: str | None = None,
        space_edge: str | list[str] | None = None,
        filter: dm.Filter | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
    ) -> ShopResultQueryAPI[T_DomainModel, T_DomainModelList]:
        """Query along the shop result edges of the benchmarking calculation input.

        Args:
            case:
            external_id_prefix:
            space:
            external_id_prefix_edge:
            space_edge:
            filter: (Advanced) Filter applied to node. If the filtering available in the
                above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            limit: Maximum number of shop result edges to return.
                Defaults to 3. Set to -1, float("inf") or None to return all items.

        Returns:
            ShopResultQueryAPI: The query API for the shop result.
        """
        from .shop_result_query import ShopResultQueryAPI

        # from is a string as we added a node query step in the __init__ method
        from_ = cast(str, self._builder.get_from())
        edge_filter = _create_edge_filter(
            dm.DirectRelationReference("power_ops_types", "ShopResults"),
            external_id_prefix=external_id_prefix_edge,
            space=space_edge,
        )
        self._builder.append(
            QueryStep(
                name=self._builder.create_name(from_),
                expression=dm.query.EdgeResultSetExpression(
                    filter=edge_filter,
                    from_=from_,
                    direction="outwards",
                ),
                max_retrieve_limit=limit,
                connection_property=ViewPropertyId(self._view_id, "shopResults"),
            )
        )

        view_id = ShopResultQueryAPI._view_id
        has_data = dm.filters.HasData(views=[view_id])
        node_filter = _create_shop_result_filter(
            view_id,
            case,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        return (ShopResultQueryAPI(
            self._client,
            self._builder,
            self._result_cls,
            self._result_list_cls,
            ViewPropertyId(self._view_id, "end_node"),
            node_filter,
            limit,
        ))

    def query(
        self,
    ) -> T_DomainModelList:
        """Execute query and return the result.

        Returns:
            The list of the source nodes of the query.

        """
        return self._query()



import math
import warnings
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Literal

from cognite.client import CogniteClient
from cognite.client.data_classes import data_modeling as dm
from cognite.client.data_classes._base import CogniteObject
from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.data_modeling.instances import Instance
from cognite.client.data_classes.data_modeling.views import ReverseDirectRelation, ViewProperty
from cognite.client.exceptions import CogniteAPIError

from cognite.powerops.client._generated.v1.data_classes._core.query.constants import (
    ACTUAL_INSTANCE_QUERY_LIMIT,
    INSTANCE_QUERY_LIMIT,
    NODE_PROPERTIES,
    NotSetSentinel,
    SelectedProperties,
)


@dataclass(frozen=True)
class ViewPropertyId(CogniteObject):
    view: dm.ViewId
    property: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> "ViewPropertyId":
        return cls(
            view=dm.ViewId.load(resource["view"]),
            property=resource["identifier"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "view": self.view.dump(camel_case=camel_case, include_type=False),
            "identifier": self.property,
        }


class QueryStep:
    """QueryStep represents a single step in a query execution.

    It is used to keep track of the state of a query step, such as the current cursor, the total number of retrieved
    instances, and the results of the query.

    Args:
        name: The unique name of the step.
        expression: The node or edge expression passed to the API.
        view_id: The view ID representing the view to query.
        max_retrieve_limit: The maximum number of instances to retrieve. Defaults to -1. If set to -1, it will
            continue to retrieve instances until there are no more instances to retrieve.
        select: The selected properties to retrieve. If not set, it will default to all properties. None indicates
            to not retrieve any properties from the view.
        raw_filter: This is the same filter as the expression, but without the HasData filter. This is used to count
            the total number of instances in the view such that the ETA can be calculated.
        connection_type: The connection type. Defaults to None.
            It is used to deal with the special case of reverse-list connections, which requires using the search
            endpoint instead of the query endpoint.
        connection_property: The property ID of the connection from the view referenced in the expression.from_
            that is used to connect to the view in this step. Defaults to None, which indicates that this is the
            root step.
        selected_properties: The user selected properties to retrieve. Defaults to None, which indicates that all
            properties should be retrieved.

    """

    def __init__(
        self,
        name: str,
        expression: dm.query.ResultSetExpression,
        view_id: dm.ViewId | None = None,
        max_retrieve_limit: int = -1,
        select: dm.query.Select | None | type[NotSetSentinel] = NotSetSentinel,
        raw_filter: dm.Filter | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        connection_property: ViewPropertyId | None = None,
        selected_properties: list[str] | None = None,
    ):
        self.name = name
        self.expression = expression
        self.view_id = view_id
        self.max_retrieve_limit = max_retrieve_limit
        self.select: dm.query.Select | None
        if select is NotSetSentinel:
            try:
                self.select = self._default_select()
            except NotImplementedError:
                raise ValueError(f"You need to provide a select to instantiate a {type(self).__name__}") from None
        else:
            self.select = select  # type: ignore[assignment]
        self.raw_filter = raw_filter
        self.connection_type = connection_type
        self.connection_property = connection_property
        self.selected_properties = selected_properties
        self._max_retrieve_batch_limit = ACTUAL_INSTANCE_QUERY_LIMIT
        self.cursor: str | None = None
        self.total_retrieved: int = 0
        self.last_batch_count: int = 0
        self.results: list[Instance] = []

    def _default_select(self) -> dm.query.Select:
        if self.view_id is None:
            return dm.query.Select()
        else:
            return dm.query.Select([dm.query.SourceSelector(self.view_id, ["*"])])

    @property
    def is_queryable(self) -> bool:
        # We cannot query across reverse-list connections
        return self.connection_type != "reverse-list"

    @property
    def from_(self) -> str | None:
        return self.expression.from_

    @property
    def is_single_direct_relation(self) -> bool:
        return isinstance(self.expression, dm.query.NodeResultSetExpression) and self.expression.through is not None

    @property
    def node_expression(self) -> dm.query.NodeResultSetExpression | None:
        if isinstance(self.expression, dm.query.NodeResultSetExpression):
            return self.expression
        return None

    @property
    def edge_expression(self) -> dm.query.EdgeResultSetExpression | None:
        if isinstance(self.expression, dm.query.EdgeResultSetExpression):
            return self.expression
        return None

    @property
    def node_results(self) -> Iterable[dm.Node]:
        return (item for item in self.results if isinstance(item, dm.Node))

    @property
    def edge_results(self) -> Iterable[dm.Edge]:
        return (item for item in self.results if isinstance(item, dm.Edge))

    def update_expression_limit(self) -> None:
        if self.is_unlimited:
            self.expression.limit = self._max_retrieve_batch_limit
        else:
            self.expression.limit = max(min(INSTANCE_QUERY_LIMIT, self.max_retrieve_limit - self.total_retrieved), 0)

    def reduce_max_batch_limit(self) -> bool:
        self._max_retrieve_batch_limit = max(1, self._max_retrieve_batch_limit // 2)
        return self._max_retrieve_batch_limit > 1

    @property
    def is_unlimited(self) -> bool:
        return self.max_retrieve_limit in {None, -1, math.inf}

    @property
    def is_finished(self) -> bool:
        return (
            (not self.is_unlimited and self.total_retrieved >= self.max_retrieve_limit)
            or self.cursor is None
            or self.last_batch_count == 0
        )

    def count_total(self, cognite_client: CogniteClient) -> float | None:
        if self.view_id is None:
            # Cannot count the total without a view
            return None
        try:
            return cognite_client.data_modeling.instances.aggregate(
                self.view_id, Count("externalId"), filter=self.raw_filter
            ).value
        except CogniteAPIError:
            return None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, from={self.from_!r}, results={len(self.results)})"


class QueryStepFactory:
    """QueryStepFactory is a factory class that creates query steps.

    Args:
        create_step_name: A function that creates a unique name for a step.
        view_id: The view ID to query. Either view_id or view must be set.
        view: The view to query. Either view_id or view must be set.
        user_selected_properties: The user selected properties to retrieve. Defaults to None, which indicates that all
            properties should be retrieved.

    """

    def __init__(
        self,
        create_step_name: Callable[[str | None], str],
        view_id: dm.ViewId | None = None,
        view: dm.View | None = None,
        user_selected_properties: SelectedProperties | None = None,
        edge_connection_property: str = "node",
    ) -> None:
        if sum(1 for x in (view_id, view) if x is not None) != 1:
            raise ValueError("Either view_id or view must be set")
        self._create_step_name = create_step_name
        if view_id is not None:
            self._view_id = view_id
        elif view is not None:
            self._view_id = view.as_id()
        else:
            raise ValueError("Either view_id or view must be set")
        self._view = view
        self._user_selected_properties = user_selected_properties
        self._root_name: str | None = None
        self._edge_connection_property = edge_connection_property

    @property
    def root_name(self) -> str:
        if self._root_name is None:
            raise ValueError("Root step is not created")
        return self._root_name

    @cached_property
    def _root_properties(self) -> list[str] | None:
        if self._user_selected_properties is None:
            return None
        root_properties: list[str] = []
        for prop in self._user_selected_properties:
            if isinstance(prop, str):
                root_properties.append(prop)
            elif isinstance(prop, dict):
                key = next(iter(prop.keys()))
                root_properties.append(key)
        return root_properties

    @cached_property
    def connection_properties(self) -> dict[str, ViewProperty]:
        output: dict[str, ViewProperty] = {}
        if self._root_properties is None or self._view is None:
            raise ValueError("View or user selected properties is required for finding" " connection properties")
        for prop in self._root_properties:
            definition = self._view.properties.get(prop)
            if not definition:
                continue
            if self._is_connection(definition):
                output[prop] = definition
        return output

    @cached_property
    def reverse_properties(self) -> dict[str, ReverseDirectRelation]:
        return {
            prop_id: prop
            for prop_id, prop in self.connection_properties.items()
            if isinstance(prop, ReverseDirectRelation)
        }

    @cached_property
    def _nested_properties_by_property(self) -> dict[str, list[str | dict[str, list[str]]]] | None:
        if self._user_selected_properties is None:
            return None
        nested_properties_by_property: dict[str, list[str | dict[str, list[str]]]] = {}
        for prop in self._user_selected_properties:
            if isinstance(prop, dict):
                key, value = next(iter(prop.items()))
                nested_properties_by_property[key] = value
        return nested_properties_by_property

    def root(
        self,
        filter: dm.Filter | None = None,
        sort: list[dm.InstanceSort] | None = None,
        limit: int | None = None,
        has_container_fields: bool = True,
    ) -> QueryStep:
        if self._root_properties:
            skip = NODE_PROPERTIES | set(self.reverse_properties.keys())
            select = self._create_select([prop for prop in self._root_properties if prop not in skip], self._view_id)
            selected_properties = self._root_properties
        else:
            select = dm.query.Select([dm.query.SourceSelector(self._view_id, ["*"])])
            selected_properties = None

        if self._root_name is not None:
            raise ValueError("Root step is already created")
        self._root_name = self._create_step_name(None)
        return QueryStep(
            self._root_name,
            dm.query.NodeResultSetExpression(
                filter=self._full_filter(filter, has_container_fields, self._view_id),
                sort=sort,
            ),
            select=select,
            selected_properties=selected_properties,
            view_id=self._view_id,
            max_retrieve_limit=-1 if limit is None else limit,
            raw_filter=filter,
        )

    def from_connection(
        self, connection_id: str, connection: ViewProperty, reverse_views: dict[dm.ViewId, dm.View]
    ) -> list[QueryStep]:
        connection_property = ViewPropertyId(self._view_id, connection_id)
        selected_properties: list[str | dict[str, list[str]]] = ["*"]
        if (nested := self._nested_properties_by_property) and connection_id in nested:
            selected_properties = nested[connection_id]

        if isinstance(connection, dm.EdgeConnection):
            return self.from_edge(connection.source, connection.direction, connection_property, selected_properties)
        elif isinstance(connection, ReverseDirectRelation):
            connection_type: Literal["reverse-list"] | None = (
                "reverse-list" if self._is_listable(connection.through, reverse_views) else None
            )
            validated = self._validate_flat_properties(selected_properties)
            return self.from_reverse_relation(
                connection.source, connection.through, connection_type, connection_property, validated
            )
        elif isinstance(connection, dm.MappedProperty) and isinstance(connection.type, dm.DirectRelation):
            validated = self._validate_flat_properties(selected_properties)
            return self.from_direct_relation(connection.source, connection_property, validated)
        else:
            warnings.warn(f"Unexpected connection type: {connection!r}", UserWarning, stacklevel=2)
        return []

    def from_direct_relation(
        self,
        source: dm.ViewId | None,
        connection_property: ViewPropertyId,
        selected_properties: list[str] | None = None,
        has_container_fields: bool = True,
    ) -> list[QueryStep]:
        if source is None:
            raise ValueError("Source view not found")
        query_properties = self._create_query_properties(selected_properties, None)
        return [
            QueryStep(
                self._create_step_name(self.root_name),
                dm.query.NodeResultSetExpression(
                    filter=self._full_filter(None, has_container_fields, source),
                    from_=self.root_name,
                    direction="outwards",
                    through=self._view_id.as_property_ref(connection_property.property),
                ),
                connection_property=connection_property,
                select=self._create_select(query_properties, source),
                selected_properties=selected_properties,
                view_id=source,
            )
        ]

    def from_edge(
        self,
        source: dm.ViewId,
        direction: Literal["outwards", "inwards"],
        connection_property: ViewPropertyId,
        selected_properties: list[str | dict[str, list[str]]] | None = None,
        include_end_node: bool = True,
        has_container_fields: bool = True,
        edge_view: dm.ViewId | None = None,
    ) -> list[QueryStep]:
        edge_name = self._create_step_name(self._root_name)
        steps = [
            QueryStep(
                edge_name,
                dm.query.EdgeResultSetExpression(
                    from_=self._root_name,
                    direction=direction,
                    chain_to="source" if direction == "outwards" else "destination",
                ),
                view_id=edge_view,
                selected_properties=[prop for prop in selected_properties or [] if isinstance(prop, str)] or None,
                connection_property=connection_property,
            )
        ]
        if not include_end_node:
            return steps

        node_properties = next(
            (prop for prop in selected_properties or [] if isinstance(prop, dict) and "node" in prop), None
        )
        selected_node_properties: list[str] | None = None
        if isinstance(node_properties, dict) and self._edge_connection_property in node_properties:
            selected_node_properties = node_properties[self._edge_connection_property]

        query_properties = self._create_query_properties(selected_node_properties, None)
        target_view = source

        step = QueryStep(
            self._create_step_name(edge_name),
            dm.query.NodeResultSetExpression(
                from_=edge_name,
                filter=self._full_filter(None, has_container_fields, target_view),
            ),
            select=self._create_select(query_properties, target_view),
            selected_properties=selected_node_properties,
            connection_property=ViewPropertyId(target_view, self._edge_connection_property),
            view_id=target_view,
        )
        steps.append(step)
        return steps

    def from_reverse_relation(
        self,
        source: dm.ViewId,
        through: dm.PropertyId,
        connection_type: Literal["reverse-list"] | None,
        connection_property: ViewPropertyId,
        selected_properties: list[str] | None = None,
        has_container_fields: bool = True,
    ) -> list[QueryStep]:
        query_properties = self._create_query_properties(selected_properties, through.property)
        other_view_id = source
        return [
            QueryStep(
                self._create_step_name(self._root_name),
                dm.query.NodeResultSetExpression(
                    from_=self._root_name,
                    direction="inwards",
                    filter=self._full_filter(None, has_container_fields, other_view_id),
                    through=other_view_id.as_property_ref(through.property),
                ),
                connection_property=connection_property,
                select=self._create_select(query_properties, other_view_id),
                selected_properties=selected_properties,
                view_id=source,
                connection_type=connection_type,
            )
        ]

    @classmethod
    def _create_query_properties(
        cls, properties: list[str] | None, connection_id: str | None = None
    ) -> list[str] | None:
        if properties is None:
            return None
        include_connection_prop = "*" not in properties
        nested_properties: list[str] = []
        for prop_id in properties:
            if prop_id in NODE_PROPERTIES:
                continue
            if prop_id == connection_id:
                include_connection_prop = False
            nested_properties.append(prop_id)

        if include_connection_prop and connection_id:
            nested_properties.append(connection_id)
        return nested_properties

    @staticmethod
    def _is_connection(definition: ViewProperty) -> bool:
        return isinstance(definition, dm.ConnectionDefinition) or (
            isinstance(definition, dm.MappedProperty) and isinstance(definition.type, dm.DirectRelation)
        )

    @staticmethod
    def _create_select(properties: list[str] | None, view_id: dm.ViewId) -> dm.query.Select:
        properties = properties or ["*"]
        return dm.query.Select([dm.query.SourceSelector(view_id, properties)])

    @staticmethod
    def _is_listable(property: dm.PropertyId, reverse_views: dict[dm.ViewId, dm.View]) -> bool:
        if isinstance(property.source, dm.ViewId):
            try:
                view = reverse_views[property.source]
            except KeyError:
                raise ValueError(f"View {property.source} not found in {reverse_views.keys()}") from None
            if property.property not in view.properties:
                raise TypeError(f"Reverse property {property.property} not found in {property.source!r}")
            reverse_prop = view.properties[property.property]
            if not (isinstance(reverse_prop, dm.MappedProperty) and isinstance(reverse_prop.type, dm.DirectRelation)):
                raise TypeError(f"Property {property.property} is not a direct relation")
            return reverse_prop.type.is_list
        else:
            raise NotImplementedError(f"Property {property.source=} is not supported")

    @staticmethod
    def _validate_flat_properties(properties: list[str | dict[str, list[str]]]) -> list[str]:
        output = []
        for prop in properties:
            if isinstance(prop, str):
                output.append(prop)
            else:
                raise ValueError(f"Direct relations do not support nested properties. Got {prop}")
        return output

    @staticmethod
    def _full_filter(filter: dm.Filter | None, has_container_fields: bool, view_id: dm.ViewId) -> dm.Filter | None:
        if has_container_fields:
            has_data = dm.filters.HasData(views=[view_id])
            return dm.filters.And(filter, has_data) if filter else has_data

        return filter



from __future__ import annotations

import datetime
from typing import (
    Annotated,
    Optional,
    Any,
    no_type_check,
)

from cognite.client import data_modeling as dm
from cognite.client.data_classes import Datapoint, Datapoints, SequenceColumnWrite, SequenceColumn
from cognite.client.data_classes import (
    TimeSeries as CogniteTimeSeries,
    Sequence as CogniteSequence,
    FileMetadata as CogniteFileMetadata,
    TimeSeriesWrite as CogniteTimeSeriesWrite,
    SequenceWrite as CogniteSequenceWrite,
    FileMetadataWrite as CogniteFileMetadataWrite,
)
from cognite.client.data_classes.sequences import ValueType
from cognite.client.utils import datetime_to_ms
from pydantic import BaseModel, BeforeValidator, model_validator, field_validator
from pydantic.alias_generators import to_camel
from pydantic.functional_serializers import PlainSerializer


TimeSeries = Annotated[
    CogniteTimeSeries,
    PlainSerializer(
        lambda v: v.dump(camel_case=True) if isinstance(v, CogniteTimeSeries) else v,
        return_type=dict,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: CogniteTimeSeries.load(v) if isinstance(v, dict) else v),
]


SequenceRead = Annotated[
    CogniteSequence,
    PlainSerializer(
        lambda v: v.dump(camel_case=True) if isinstance(v, CogniteSequence) else v,
        return_type=dict,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: CogniteSequence.load(v) if isinstance(v, dict) else v),
]


FileMetadata = Annotated[
    CogniteFileMetadata,
    PlainSerializer(
        lambda v: v.dump(camel_case=True) if isinstance(v, CogniteFileMetadata) else v,
        return_type=dict,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: CogniteFileMetadata.load(v) if isinstance(v, dict) else v),
]


TimeSeriesWrite = Annotated[
    CogniteTimeSeriesWrite,
    PlainSerializer(
        lambda v: v.dump(camel_case=True) if isinstance(v, CogniteTimeSeriesWrite) else v,
        return_type=dict,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: CogniteTimeSeriesWrite.load(v) if isinstance(v, dict) else v),
]


SequenceWrite = Annotated[
    CogniteSequenceWrite,
    PlainSerializer(
        lambda v: v.dump(camel_case=True) if isinstance(v, CogniteSequenceWrite) else v,
        return_type=dict,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: CogniteSequenceWrite.load(v) if isinstance(v, dict) else v),
]

FileMetadataWrite = Annotated[
    CogniteFileMetadataWrite,
    PlainSerializer(
        lambda v: v.dump(camel_case=True) if isinstance(v, CogniteFileMetadataWrite) else v,
        return_type=dict,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: CogniteFileMetadataWrite.load(v) if isinstance(v, dict) else v),
]


class GraphQLExternal(BaseModel, arbitrary_types_allowed=True, populate_by_name=True, alias_generator=to_camel): ...


class TimeSeriesGraphQL(GraphQLExternal):
    id: Optional[int] = None
    external_id: Optional[str] = None
    instance_id: Optional[dm.NodeId] = None
    name: Optional[str] = None
    is_string: Optional[bool] = None
    metadata: Optional[dict[str, str]] = None
    unit: Optional[str] = None
    unit_external_id: Optional[str] = None
    asset_id: Optional[int] = None
    is_step: Optional[bool] = None
    description: Optional[str] = None
    security_categories: Optional[list[int]] = None
    data_set_id: Optional[int] = None
    created_time: Optional[int] = None
    last_updated_time: Optional[int] = None
    data: Optional[Datapoints] = None
    latest: Optional[Datapoint] = None

    @model_validator(mode="before")
    def parse_datapoints(cls, data: Any) -> Any:
        if isinstance(data, dict) and "getDataPoints" in data:
            datapoints = data.pop("getDataPoints")
            if "items" in datapoints:
                for item in datapoints["items"]:
                    # The Datapoints expects the timestamp to be in milliseconds
                    item["timestamp"] = datetime_to_ms(
                        datetime.datetime.fromisoformat(item["timestamp"].replace("Z", "+00:00"))
                    )
                data["datapoints"] = datapoints["items"]
                data["data"] = Datapoints.load(data)
        if isinstance(data, dict) and "getLatestDataPoint" in data:
            latest = data.pop("getLatestDataPoint")
            if "items" in latest:
                latest["items"][0]["timestamp"] = datetime_to_ms(
                    datetime.datetime.fromisoformat(latest["items"][0]["timestamp"].replace("Z", "+00:00"))
                )
                data["latest"] = Datapoint.load(latest["items"][0])
        return data

    def as_write(self) -> CogniteTimeSeriesWrite:
        return CogniteTimeSeriesWrite(
            external_id=self.external_id,
            name=self.name,
            is_string=self.is_string,
            metadata=self.metadata,
            unit=self.unit,
            unit_external_id=self.unit_external_id,
            asset_id=self.asset_id,
            is_step=self.is_step,
            description=self.description,
        )

    def as_read(self) -> CogniteTimeSeries:
        return CogniteTimeSeries(
            id=self.id,
            external_id=self.external_id,
            instance_id=self.instance_id,
            name=self.name,
            is_string=self.is_string,
            metadata=self.metadata,
            unit=self.unit,
            unit_external_id=self.unit_external_id,
            asset_id=self.asset_id,
            is_step=self.is_step,
            description=self.description,
            security_categories=self.security_categories,
        )


class FileMetadataGraphQL(GraphQLExternal):
    external_id: Optional[str] = None
    name: Optional[str] = None
    source: Optional[str] = None
    mime_type: Optional[str] = None
    metadata: Optional[dict[str, str]] = None
    directory: Optional[str] = None
    asset_ids: Optional[list[int]] = None
    data_set_id: Optional[int] = None
    labels: Optional[list[dict]] = None
    geo_location: Optional[dict] = None
    source_created_time: Optional[int] = None
    source_modified_time: Optional[int] = None
    security_categories: Optional[list[int]] = None
    id: Optional[int] = None
    uploaded: Optional[bool] = None
    uploaded_time: Optional[int] = None
    created_time: Optional[int] = None
    last_updated_time: Optional[int] = None

    @no_type_check
    def as_write(self) -> CogniteFileMetadataWrite:
        return CogniteFileMetadataWrite(
            external_id=self.external_id,
            name=self.name,
            source=self.source,
            mime_type=self.mime_type,
            metadata=self.metadata,
            directory=self.directory,
            asset_ids=self.asset_ids,
            data_set_id=self.data_set_id,
            labels=self.labels,
            geo_location=self.geo_location,
            source_created_time=self.source_created_time,
            source_modified_time=self.source_modified_time,
            security_categories=self.security_categories,
        )

    @no_type_check
    def as_read(self) -> CogniteFileMetadata:
        return CogniteFileMetadata(
            external_id=self.external_id,
            name=self.name,
            source=self.source,
            mime_type=self.mime_type,
            metadata=self.metadata,
            directory=self.directory,
            asset_ids=self.asset_ids,
            data_set_id=self.data_set_id,
            labels=self.labels,
            geo_location=self.geo_location,
            source_created_time=self.source_created_time,
            source_modified_time=self.source_modified_time,
            security_categories=self.security_categories,
            id=self.id,
            uploaded=self.uploaded,
            uploaded_time=self.uploaded_time,
            created_time=self.created_time,
            last_updated_time=self.last_updated_time,
        )


class SequenceColumnGraphQL(GraphQLExternal):
    external_id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    value_type: Optional[ValueType] = None
    metadata: Optional[dict[str, str]] = None
    created_time: Optional[int] = None
    last_updated_time: Optional[int] = None

    @field_validator("value_type", mode="before")
    def title_value_type(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.title()
        return value

    @no_type_check
    def as_write(self) -> SequenceColumnWrite:
        if self.value_type is None:
            raise ValueError("value_type is required")
        return SequenceColumnWrite(
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            value_type=self.value_type,
            metadata=self.metadata,
        )

    @no_type_check
    def as_read(self) -> SequenceColumn:
        if self.value_type is None:
            raise ValueError("value_type is required")
        return SequenceColumn(
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            value_type=self.value_type,
            metadata=self.metadata,
            created_time=self.created_time,
            last_updated_time=self.last_updated_time,
        )


class SequenceGraphQL(GraphQLExternal):
    id: Optional[int] = None
    name: Optional[str] = None
    description: Optional[str] = None
    asset_id: Optional[int] = None
    external_id: Optional[str] = None
    metadata: Optional[dict[str, str]] = None
    columns: Optional[list[SequenceColumnGraphQL]] = None
    created_time: Optional[int] = None
    last_updated_time: Optional[int] = None
    data_set_id: Optional[int] = None

    @no_type_check
    def as_write(self) -> CogniteSequenceWrite:
        return CogniteSequenceWrite(
            name=self.name,
            description=self.description,
            asset_id=self.asset_id,
            external_id=self.external_id,
            metadata=self.metadata,
            columns=[col.as_write() for col in self.columns or []] if self.columns else None,
            data_set_id=self.data_set_id,
        )

    @no_type_check
    def as_read(self) -> CogniteSequence:
        return CogniteSequence(
            id=self.id,
            name=self.name,
            description=self.description,
            asset_id=self.asset_id,
            external_id=self.external_id,
            metadata=self.metadata,
            columns=[col.as_read() for col in self.columns or []] if self.columns else None,
            created_time=self.created_time,
            last_updated_time=self.last_updated_time,
            data_set_id=self.data_set_id,
        )



from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import Any, ClassVar, Literal, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList, InstanceSort

from cognite.powerops.client._generated.v1._api._core import (
    DEFAULT_LIMIT_READ,
    instantiate_classes,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
)
from cognite.powerops.client._generated.v1.data_classes._core import (
    DEFAULT_INSTANCE_SPACE,
    DEFAULT_QUERY_LIMIT,
    QueryStepFactory,
    QueryBuilder,
    QueryUnpacker,
    ViewPropertyId,
)
from cognite.powerops.client._generated.v1.data_classes._generator_efficiency_curve import (
    GeneratorEfficiencyCurveQuery,
    _GENERATOREFFICIENCYCURVE_PROPERTIES_BY_FIELD,
    _create_generator_efficiency_curve_filter,
)
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    GeneratorEfficiencyCurve,
    GeneratorEfficiencyCurveWrite,
    GeneratorEfficiencyCurveFields,
    GeneratorEfficiencyCurveList,
    GeneratorEfficiencyCurveWriteList,
    GeneratorEfficiencyCurveTextFields,
)
from cognite.powerops.client._generated.v1._api.generator_efficiency_curve_query import GeneratorEfficiencyCurveQueryAPI


class GeneratorEfficiencyCurveAPI(NodeAPI[GeneratorEfficiencyCurve, GeneratorEfficiencyCurveWrite, GeneratorEfficiencyCurveList, GeneratorEfficiencyCurveWriteList]):
    _view_id = dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1")
    _properties_by_field: ClassVar[dict[str, str]] = _GENERATOREFFICIENCYCURVE_PROPERTIES_BY_FIELD
    _class_type = GeneratorEfficiencyCurve
    _class_list = GeneratorEfficiencyCurveList
    _class_write_list = GeneratorEfficiencyCurveWriteList

    def __init__(self, client: CogniteClient):
        super().__init__(client=client)


    def __call__(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> GeneratorEfficiencyCurveQueryAPI[GeneratorEfficiencyCurve, GeneratorEfficiencyCurveList]:
        """Query starting at generator efficiency curves.

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generator efficiency curves to return. Defaults to 25.
                Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write
                your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for generator efficiency curves.

        """
        warnings.warn(
            "This method is deprecated and will soon be removed. "
            "Use the .select() method instead.",
            UserWarning,
            stacklevel=2,
        )
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_generator_efficiency_curve_filter(
            self._view_id,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        return GeneratorEfficiencyCurveQueryAPI(
            self._client, QueryBuilder(), self._class_type, self._class_list, None, filter_, limit
        )

    def apply(
        self,
        generator_efficiency_curve: GeneratorEfficiencyCurveWrite | Sequence[GeneratorEfficiencyCurveWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) generator efficiency curves.

        Args:
            generator_efficiency_curve: Generator efficiency curve or
                sequence of generator efficiency curves to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and
                existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)?
                Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None.
                However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new generator_efficiency_curve:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import GeneratorEfficiencyCurveWrite
                >>> client = PowerOpsModelsV1Client()
                >>> generator_efficiency_curve = GeneratorEfficiencyCurveWrite(
                ...     external_id="my_generator_efficiency_curve", ...
                ... )
                >>> result = client.generator_efficiency_curve.apply(generator_efficiency_curve)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.generator_efficiency_curve.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(generator_efficiency_curve, replace, write_none)

    def delete(self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> dm.InstancesDeleteResult:
        """Delete one or more generator efficiency curve.

        Args:
            external_id: External id of the generator efficiency curve to delete.
            space: The space where all the generator efficiency curve are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete generator_efficiency_curve by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.generator_efficiency_curve.delete("my_generator_efficiency_curve")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.generator_efficiency_curve.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(
        self,
        external_id: str | dm.NodeId | tuple[str, str],
        space: str = DEFAULT_INSTANCE_SPACE,
    ) -> GeneratorEfficiencyCurve | None: ...

    @overload
    def retrieve(
        self,
        external_id: SequenceNotStr[str | dm.NodeId | tuple[str, str]],
        space: str = DEFAULT_INSTANCE_SPACE,
    ) -> GeneratorEfficiencyCurveList: ...

    def retrieve(
        self,
        external_id: str | dm.NodeId | tuple[str, str] | SequenceNotStr[str | dm.NodeId | tuple[str, str]],
        space: str = DEFAULT_INSTANCE_SPACE,
    ) -> GeneratorEfficiencyCurve | GeneratorEfficiencyCurveList | None:
        """Retrieve one or more generator efficiency curves by id(s).

        Args:
            external_id: External id or list of external ids of the generator efficiency curves.
            space: The space where all the generator efficiency curves are located.

        Returns:
            The requested generator efficiency curves.

        Examples:

            Retrieve generator_efficiency_curve by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> generator_efficiency_curve = client.generator_efficiency_curve.retrieve(
                ...     "my_generator_efficiency_curve"
                ... )

        """
        return self._retrieve(
            external_id,
            space,
        )

    def search(
        self,
        query: str,
        properties: GeneratorEfficiencyCurveTextFields | SequenceNotStr[GeneratorEfficiencyCurveTextFields] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        sort_by: GeneratorEfficiencyCurveFields | SequenceNotStr[GeneratorEfficiencyCurveFields] | None = None,
        direction: Literal["ascending", "descending"] = "ascending",
        sort: InstanceSort | list[InstanceSort] | None = None,
    ) -> GeneratorEfficiencyCurveList:
        """Search generator efficiency curves

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generator efficiency curves to return. Defaults to 25.
                Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient,
                you can write your own filtering which will be ANDed with the filter above.
            sort_by: The property to sort by.
            direction: The direction to sort by, either 'ascending' or 'descending'.
            sort: (Advanced) If sort_by and direction are not sufficient, you can write your own sorting.
                This will override the sort_by and direction. This allows you to sort by multiple fields and
                specify the direction for each field as well as how to handle null values.

        Returns:
            Search results generator efficiency curves matching the query.

        Examples:

           Search for 'my_generator_efficiency_curve' in all text properties:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> generator_efficiency_curves = client.generator_efficiency_curve.search(
                ...     'my_generator_efficiency_curve'
                ... )

        """
        filter_ = _create_generator_efficiency_curve_filter(
            self._view_id,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(
            query=query,
            properties=properties,
            filter_=filter_,
            limit=limit,
            sort_by=sort_by,  # type: ignore[arg-type]
            direction=direction,
            sort=sort,
        )

    @overload
    def aggregate(
        self,
        aggregate: Aggregations | dm.aggregations.MetricAggregation,
        group_by: None = None,
        property: GeneratorEfficiencyCurveFields | SequenceNotStr[GeneratorEfficiencyCurveFields] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.AggregatedNumberedValue: ...

    @overload
    def aggregate(
        self,
        aggregate: SequenceNotStr[Aggregations | dm.aggregations.MetricAggregation],
        group_by: None = None,
        property: GeneratorEfficiencyCurveFields | SequenceNotStr[GeneratorEfficiencyCurveFields] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | SequenceNotStr[Aggregations | dm.aggregations.MetricAggregation],
        group_by: GeneratorEfficiencyCurveFields | SequenceNotStr[GeneratorEfficiencyCurveFields],
        property: GeneratorEfficiencyCurveFields | SequenceNotStr[GeneratorEfficiencyCurveFields] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | SequenceNotStr[Aggregations | dm.aggregations.MetricAggregation],
        group_by: GeneratorEfficiencyCurveFields | SequenceNotStr[GeneratorEfficiencyCurveFields] | None = None,
        property: GeneratorEfficiencyCurveFields | SequenceNotStr[GeneratorEfficiencyCurveFields] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> (
        dm.aggregations.AggregatedNumberedValue
        | list[dm.aggregations.AggregatedNumberedValue]
        | InstanceAggregationResultList
    ):
        """Aggregate data across generator efficiency curves

        Args:
            aggregate: The aggregation to perform.
            group_by: The property to group by when doing the aggregation.
            property: The property to perform aggregation on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generator efficiency curves to return. Defaults to 25.
                Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write
                your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count generator efficiency curves in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.generator_efficiency_curve.aggregate("count", space="my_space")

        """

        filter_ = _create_generator_efficiency_curve_filter(
            self._view_id,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            aggregate=aggregate,
            group_by=group_by,  # type: ignore[arg-type]
            properties=property,  # type: ignore[arg-type]
            query=None,
            search_properties=None,
            limit=limit,
            filter=filter_,
        )

    def histogram(
        self,
        property: GeneratorEfficiencyCurveFields,
        interval: float,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for generator efficiency curves

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generator efficiency curves to return.
                Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient,
                you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_generator_efficiency_curve_filter(
            self._view_id,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            property,
            interval,
            None,
            None,
            limit,
            filter_,
        )

    def select(self) -> GeneratorEfficiencyCurveQuery:
        """Start selecting from generator efficiency curves."""
        return GeneratorEfficiencyCurveQuery(self._client)

    def _query(
        self,
        filter_: dm.Filter | None,
        limit: int,
        retrieve_connections: Literal["skip", "identifier", "full"],
        sort: list[InstanceSort] | None = None,
    ) -> list[dict[str, Any]]:
        builder = QueryBuilder()
        factory = QueryStepFactory(builder.create_name, view_id=self._view_id, edge_connection_property="end_node")
        builder.append(factory.root(
            filter=filter_,
            sort=sort,
            limit=limit,
            has_container_fields=True,
        ))
        unpack_edges: Literal["skip", "identifier"] = "identifier" if retrieve_connections == "identifier" else "skip"
        builder.execute_query(self._client, remove_not_connected=True if unpack_edges == "skip" else False)
        return QueryUnpacker(builder, edges=unpack_edges).unpack()


    def list(
        self,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        sort_by: GeneratorEfficiencyCurveFields | Sequence[GeneratorEfficiencyCurveFields] | None = None,
        direction: Literal["ascending", "descending"] = "ascending",
        sort: InstanceSort | list[InstanceSort] | None = None,
    ) -> GeneratorEfficiencyCurveList:
        """List/filter generator efficiency curves

        Args:
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generator efficiency curves to return.
                Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient,
                you can write your own filtering which will be ANDed with the filter above.
            sort_by: The property to sort by.
            direction: The direction to sort by, either 'ascending' or 'descending'.
            sort: (Advanced) If sort_by and direction are not sufficient, you can write your own sorting.
                This will override the sort_by and direction. This allowos you to sort by multiple fields and
                specify the direction for each field as well as how to handle null values.

        Returns:
            List of requested generator efficiency curves

        Examples:

            List generator efficiency curves and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> generator_efficiency_curves = client.generator_efficiency_curve.list(limit=5)

        """
        filter_ = _create_generator_efficiency_curve_filter(
            self._view_id,
            external_id_prefix,
            space,
            filter,
        )
        sort_input =  self._create_sort(sort_by, direction, sort)  # type: ignore[arg-type]
        return self._list(limit=limit,  filter=filter_, sort=sort_input)
